/*
 * @(#)SelectCommand.java 5.1
 *
 */

package CH.ifa.draw.figures;

import java.util.*;
import CH.ifa.draw.util.Command;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.standard.*;

/**
 * Command to select certain Figures of the current drawing.
 *
 * @see SelectionTool
 */
public  class SelectCommand extends Command {

   /** Value for the mode parameter of some constructors:
    * The select command behaves in its 'classical' way,
    * it clears any previous selection.
    * This mode is the default if constructors without
    * a mode parameter are used.
    **/
    public static final int SELECT = 0;

   /** Value for the mode parameter of some constructors:
    * The select command adds the specified figures to
    * the selection.
    **/
    public static final int ADD    = 1;

   /** Value for the mode parameter of some constructors:
    * The select command removes the specified figures from
    * the selection.
    **/
    public static final int REMOVE = 2;

   /** Value for the mode parameter of some constructors:
    * The select command restricts the selection to the specified figures.
    **/
    public static final int RESTRICT = 3;

    protected DrawingEditor fEditor;
    protected FigureFilter fFilter;
    protected int selectMode;

   /**
    * Constructs a select all command.
    * @param name the command name
    * @param view the target editor
    */
    public SelectCommand(String name, DrawingEditor editor) {
        this(name, editor, Figure.class, null, SELECT);
    }

   /**
    * Constructs a select command for certain figures.
    * @param name the command name
    * @param view the target editor
    * @param figureClass the class of figures to be selected
    */
    public SelectCommand(String name, DrawingEditor editor,
			 Class figureClass) {
        this(name, editor, figureClass, SELECT);
    }

   /**
    * Constructs a select/add to selection/remove from selection
    * command for certain figures.
    * @param name the command name
    * @param view the target editor
    * @param figureClass the class of figures to be selected
    * @param selectMode one of the constants SELECT, ADD, REMOVE
    *           or RESTRICT defined in this class
    */
    public SelectCommand(String name, DrawingEditor editor,
			 Class figureClass, int selectMode) {
        this(name, editor, figureClass, null, selectMode);
    }

   /**
    * Constructs a select command for certain figures.
    * @param name the command name
    * @param view the target editor
    * @param figureClass the class of figures to be selected
    * @param parentClass also check this class for the
    *           parent of the child figures to be selected.
    *           if parentClass != null, figureClass has to be
    *           a subclass of ChildFigure and parentClass should
    *           be a subclass of ParentFigure.
    */
    public SelectCommand(String name, DrawingEditor editor,
			 final Class figureClass, final Class parentClass) {
        this(name, editor, figureClass, parentClass, SELECT);
    }

   /**
    * Constructs a select/add to selection/remove from selection
    * command for certain figures.
    * @param name the command name
    * @param view the target editor
    * @param figureClass the class of figures to be selected
    * @param parentClass also check this class for the
    *           parent of the child figures to be selected.
    *           if parentClass != null, figureClass has to be
    *           a subclass of ChildFigure and parentClass should
    *           be a subclass of ParentFigure.
    * @param selectMode one of the constants SELECT, ADD, REMOVE
    *           or RESTRICT defined in this class
    */
    public SelectCommand(String name, DrawingEditor editor,
			 final Class figureClass, final Class parentClass,
			 int selectMode) {
        this(name, editor, new FigureFilter() {
	    public boolean isUsed(Figure fig) {
		return figureClass.isInstance(fig) &&
		    (parentClass==null ||
		     parentClass.isInstance(((ChildFigure)fig).parent()));
	    }
	}, selectMode);
    }

   /**
    * Constructs a select command for certain figures.
    * @param name the command name
    * @param view the target editor
    * @param filter the filter object that decides whether a figure
    *           will be selected
    */
    public SelectCommand(String name, DrawingEditor editor,
			 FigureFilter filter) {
        this(name, editor, filter, SELECT);
    }

   /**
    * Constructs a select/add to selection/remove from selection
    * command for certain figures.
    * @param name the command name
    * @param view the target editor
    * @param filter the filter object that decides whether a figure
    *           will be selected
    * @param selectMode one of the constants SELECT, ADD, REMOVE
    *           or RESTRICT defined in this class
    */
    public SelectCommand(String name, DrawingEditor editor,
			 FigureFilter filter, int selectMode) {
        super(name);
        fEditor = editor;
	fFilter = filter;
	this.selectMode = selectMode;
    }



    public void execute() {
        DrawingView view = fEditor.view();
        Drawing drawing = view.drawing();

	if (selectMode == SELECT) view.clearSelection();

        FigureEnumeration figenum=drawing.figures();
        while (figenum.hasMoreElements()) {
	    Figure fig=figenum.nextFigure();
	    if (fFilter.isUsed(fig)) {
	        if (selectMode == REMOVE)
		    view.removeFromSelection(fig);
		else if (selectMode != RESTRICT)
		    view.addToSelection(fig);
	    } else {
	        if (selectMode == RESTRICT)
		    view.removeFromSelection(fig);
	    }
        }
        view.checkDamage();
    }

}


