/*
 * @(#)TextFigure.java 5.1
 *
 */

package CH.ifa.draw.figures;

import java.util.*;
import java.awt.*;
import java.io.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.standard.*;
import CH.ifa.draw.util.*;
import CH.ifa.draw.application.*;

/**
 * A text figure.
 *
 * @see TextTool
 */
public  class TextFigure
        extends AttributeFigure
        implements ChildFigure, TextHolder {

    private int               fOriginX;
    private int               fOriginY;

    // cache of the TextFigure's size
    transient private boolean fSizeIsDirty = true;
    transient private int     fWidth;
    transient private int     fHeight;

    private String  fText;
    private String[] fLines;

    private Font    fFont = null;
    private static String fgCurrentFontName  = "Helvetica";
    private static int    fgCurrentFontSize  = 12;
    private static int    fgCurrentFontStyle = Font.PLAIN;
    private String  fCurrentFontName  = fgCurrentFontName;
    private int     fCurrentFontSize;
    private int     fCurrentFontStyle = fgCurrentFontStyle;
    private boolean fIsReadOnly;
    private boolean fCanBeConnected=true;

    private ParentFigure  fParent = null;
    private OffsetLocator fLocator = null;

    /*
     * Serialization support.
     */
    private static final long serialVersionUID = 4599820785949456124L;
    private int textFigureSerializedDataVersion = 1;

    static {
	String fontSizeStr=System.getProperty("de.renew.defaultFontSize");
	if (fontSizeStr!=null) {
	    try {
		fgCurrentFontSize=Integer.parseInt(fontSizeStr);
		System.out.println("Setting default font size to "+fgCurrentFontSize+" pt.");
	    } catch (NumberFormatException e) {
		System.out.println("Failed to set default font size to \""+fontSizeStr+"\": wrong number format.");
	    }
	}
    }

    public TextFigure() {
       this("");
    }

    public TextFigure(boolean canBeConnected) {
	this("");
	fCanBeConnected=canBeConnected;
    }

    public TextFigure(String text) {
	fCurrentFontSize  = fgCurrentFontSize;
        fOriginX = 0;
        fOriginY = 0;
        setAttribute("FillColor", ColorMap.color("None"));
        setAttribute("FrameColor", ColorMap.color("None"));
	internalSetText(text);
        fSizeIsDirty = true;
    }

    public TextFigure(String text, boolean isReadOnly) {
       this(text);
       setReadOnly(isReadOnly);
    }

    public void moveBy(int x, int y) {
        willChange();
        basicMoveBy(x, y);
        if (fLocator != null)
            fLocator.moveBy(x, y);
        changed();
    }

    protected void basicMoveBy(int x, int y) {
        fOriginX += x;
        fOriginY += y;
    }

    public void basicDisplayBox(Point newOrigin, Point newCorner) {
        fOriginX = newOrigin.x;
        fOriginY = newOrigin.y;
    }

    public Point getOrigin() {
      return new Point(fOriginX, fOriginY);
    }

    public Rectangle displayBox() {
      textExtent();
      return new Rectangle(fOriginX, fOriginY, fWidth, fHeight);
    }

    public Rectangle textDisplayBox() {
        return displayBox();
    }

  protected Dimension[] getLineDimensions() {
    FontMetrics metrics =
      Toolkit.getDefaultToolkit().getFontMetrics(getFont());
    String[] lines=getLines();
    Dimension[] dimensions=new Dimension[lines.length];
    for (int i=0;i<lines.length;i++) {
      dimensions[i]=oneLine(null,metrics,lines[i],0,0);
    }
    return dimensions;
  }

    /**
     * Checks if a point is inside the figure.
     */
    public boolean containsPoint(int x, int y) {
      if (super.containsPoint(x,y)) {
	if (!ColorMap.isTransparent(getFrameColor()))
	    return true; // regard as box!
	// check if the corresponding line is long enough:
        Dimension[] dimensions=getLineDimensions();
	int h = fOriginY;
	for (int i=0;i<dimensions.length;i++) {
	  h += dimensions[i].height;
	  if (y<h)
	    return x-fOriginX <= dimensions[i].width+1;
	}
      }
      return false;
    }

    /**
     * Tests whether this figure is read only.
     */
    public boolean readOnly() {
        return fIsReadOnly;
    }

    /**
     * Sets the read only status of the text figure.
     */
    public void setReadOnly(boolean isReadOnly) {
        fIsReadOnly = isReadOnly;
    }

    /**
     * Gets the font.
     */
    public Font getFont() {
      if (fFont == null)
	  fFont = Fontkit.getFont(fCurrentFontName, fCurrentFontStyle,
				  fCurrentFontSize);
      return fFont;
    }

    /**
     * Sets the font.
     */
    public void setFont(Font newFont) {
        willChange();
        fFont = newFont;
	fCurrentFontName = fFont.getName();
	fCurrentFontStyle = fFont.getStyle();
	fCurrentFontSize = fFont.getSize();
        markDirty();
        changed();
    }

    /**
     * Updates the location whenever the figure changes itself.
     */
    public void changed() {
        super.changed();
        updateLocation();
    }

    /**
     * A text figure understands the "FontSize", "FontStyle", and "FontName"
     * attributes.
     */
    public Object getAttribute(String name) {
        if (name.equals("FontSize"))
            return new Integer(fCurrentFontSize);
        if (name.equals("FontStyle"))
            return new Integer(fCurrentFontStyle);
        if (name.equals("FontName"))
            return fCurrentFontName;
        return super.getAttribute(name);
    }

    /**
     * A text figure understands the "FontSize", "FontStyle", and "FontName"
     * attributes.
     */
    public void setAttribute(String name, Object value) {
      if (name.equals("FontSize") || name.equals("FontStyle")
	  || name.equals("FontName")) {
	willChange();
        if (name.equals("FontSize")) {
            fCurrentFontSize = ((Integer)value).intValue();
        }
        else if (name.equals("FontStyle")) {
            int s = ((Integer)value).intValue();
            if (s == Font.PLAIN)
                fCurrentFontStyle = Font.PLAIN;
            else
                fCurrentFontStyle = fCurrentFontStyle ^ s;
        }
        else if (name.equals("FontName")) {
            fCurrentFontName = (String)value;
        }
	fFont = null;
        markDirty();
        changed();

        markDirty();
        changed();
      } else {
	super.setAttribute(name, value);
      }
    }

    /**
     * Gets the text shown by the text figure.
     */
    public String getText() {
        return fText;
    }

  /**
   * Get the text shown by the text figure, but splitted into
   * an array of individual lines.
   */
  public String[] getLines() {
    return fLines;
  }

    /**
     * Sets the text shown by the text figure.
     */
    public void setText(String newText) {
        if (!newText.equals(fText)) {
            willChange();
	    basicSetText(newText);
            changed();
        }
    }

    protected void basicSetText(String newText) {
      internalSetText(newText);
      markDirty();
    }

  protected void internalSetText(String newText) {
    fText = newText;
    fLines = splitTextLines(newText);
  }

  private static String[] splitTextLines(String text) {
    Vector vector=new Vector();
    StringTokenizer lines = new StringTokenizer(text,"\n",true);
    boolean lastLineTerminated=true;
    while (lines.hasMoreElements()) {
      String line=lines.nextToken();
      if ("\n".equals(line)) {
	line="";
	lastLineTerminated=true;
      } else if (lines.hasMoreElements()) {
	// it was not only a separator, skip its separator:
	lines.nextToken();
	lastLineTerminated=true;
      } else {
	lastLineTerminated=false;
      }
      vector.addElement(line);
    }
    if (lastLineTerminated) {
      vector.addElement("");
    }
    String[] arr=new String[vector.size()];
    vector.copyInto(arr);
    return arr;
  }

    /**
     * Tests whether the figure accepts typing.
     */
    public boolean acceptsTyping() {
        return !fIsReadOnly;
    }

    public void draw(Graphics g) {
	if (isVisible()) {
	    super.draw(g);
	    FontMetrics metrics = g.getFontMetrics(getFont());
	    Color fillColor=getFillColor();
	    String[] lines=getLines();
	    if (ColorMap.isTransparent(getFrameColor()) &&
		!ColorMap.isTransparent(fillColor)) {
		// fill each line individually:
		g.setColor(fillColor);
		int y = fOriginY;
		for (int i=0;i<lines.length;i++) {
		  Dimension dim=oneLine(null,metrics,lines[i],fOriginX,y);
		  g.fillRect(fOriginX,y,dim.width,dim.height);
		  y += dim.height;
		}
	    }
	    Color textColor=(Color)getAttribute("TextColor");
	    if (!ColorMap.isTransparent(textColor)) {
		g.setColor(textColor);
		g.setFont(getFont());
		int y = fOriginY;
		for (int i=0;i<lines.length;i++) {
		  Dimension dim=oneLine(g,metrics,lines[i],fOriginX,y);
		  y += dim.height;
		}
	    }
	}
    }

    protected Dimension oneLine(Graphics g, FontMetrics metrics,
				String line, int x, int y) {
	int w=metrics.stringWidth(line),
	    h=metrics.getHeight();
	if (g!=null) {
	  // This was originally
	  //   g.drawString(line, x, y);
	  // but there is a bug in jdk1.1.7 drawString:
	  // String is not properly converted to different encodings.  
	  //
	  // However there is another bug in the Sun JDK:
	  // drawBytes does not respect the font before drawString
	  // is called once.
	  //g.drawString(" ",-10000,-10000);

	  // Now use drawBytes instead of drawString!
	  y += metrics.getAscent();
	  g.drawString(line,x,y);
	  //g.drawBytes(line.getBytes(),0,line.length(),x,y);
	  //if ((g.getFont().getStyle() & ExtendedFont.UNDERLINED) != 0) {
	  // g.drawLine(x, y+2, x+w, y+2);
	  //}
	}
	return new Dimension(w,h);
    }

    public void drawBackground(Graphics g) {
      if (!ColorMap.isTransparent(getFrameColor())) {
        Rectangle r = displayBox();
        g.fillRect(r.x, r.y, r.width, r.height);
      }
    }

    public void drawFrame(Graphics g) {
        Rectangle r = displayBox();
        g.drawRect(r.x, r.y, r.width-1, r.height-1);
    }

    private Dimension textExtent() {
      if (fSizeIsDirty) {
	fWidth = 0;
	fHeight = 0;

	Dimension[] dimensions=getLineDimensions();
	for (int i=0;i<dimensions.length;i++) {
	  fWidth = Math.max(fWidth,dimensions[i].width);
	  fHeight += dimensions[i].height;
	}

        fSizeIsDirty = false;
      }
      return new Dimension(fWidth, fHeight);
    }

    protected void markDirty() {
        fSizeIsDirty = true;
    }

    /**
     * Gets the number of rows and columns to be overlaid when the figure
     * is edited.
     */
    public Dimension overlayRowsAndColumns() {
      int columns = 20;
      String[] lines=getLines();
      for (int i=0;i<lines.length;i++) {
	columns = Math.max(columns,lines[i].length() + 3);
      }
      int rows=Math.max(1,lines.length);
      
      return new Dimension(columns, rows);
    }

    public Vector handles() {
      Vector handles=new Vector();
      handles.addElement(new NullHandle(this, RelativeLocator.northWest()));
      handles.addElement(new NullHandle(this, RelativeLocator.northEast()));
      handles.addElement(new NullHandle(this, RelativeLocator.southEast()));
      handles.addElement(new FontSizeHandle(this,RelativeLocator.southWest()));
      return handles;
    }

    public void write(StorableOutput dw) {
        super.write(dw);
        dw.writeInt(fOriginX);
        dw.writeInt(fOriginY);
        dw.writeString(fText);
        dw.writeString(fCurrentFontName);
        dw.writeInt(fCurrentFontStyle);
        dw.writeInt(fCurrentFontSize);
        dw.writeBoolean(fIsReadOnly);
        dw.writeStorable(fParent);
        dw.writeStorable(fLocator);
    }

    public void read(StorableInput dr) throws IOException {
        super.read(dr);
        fOriginX = dr.readInt();
        fOriginY = dr.readInt();
        String text = dr.readString();
	setText(text);
	fCurrentFontName = dr.readString();
	fCurrentFontStyle = dr.readInt();
	fCurrentFontSize = dr.readInt();
        fFont = null;
        fIsReadOnly = dr.readBoolean();

        fParent = (ParentFigure)dr.readStorable();
        fLocator = (OffsetLocator)dr.readStorable();
        if (fParent != null)
            fParent.addChild(this);
	if (System.getProperty("de.renew.noGraphics") != null)
	  fSizeIsDirty = false;
    }

    private void readObject(ObjectInputStream s)
        throws ClassNotFoundException, IOException {

        s.defaultReadObject();

        //if (fParent != null) {
        //    fParent.addChild(this);
	//}
        markDirty();
    }

    public boolean canBeParent(ParentFigure figure) {
      if (!fCanBeConnected)
	  return figure==null;
      while (figure != null) {
	if (figure == this)
	  return false;     // cyclic!!!
        if (figure instanceof ChildFigure)
	  figure = ((ChildFigure)figure).parent();
	else
	  break;
      }
      return true;
    }

    public boolean setParent(ParentFigure figure) {
        if (!canBeParent(figure))
	  return false;
        if (fParent != null) {
           fParent.removeChild(this);
        }
        fParent = figure;
	if (fParent == null) {
          fLocator = null;
        } else {
	  fParent.addChild(this);
          if (fLocator != null)
             fLocator.setBase(fParent.connectedTextLocator(this));
          else
	     fLocator = new OffsetLocator(fParent.connectedTextLocator(this));
	  updateLocation();
	}
	return true;

    }

    public ParentFigure parent() {
        return fParent;
    }

    /**
     * Updates the location relative to the connected figure.
     * The TextFigure is centered around the located point.
     */
    public void updateLocation() {
        if (fLocator != null) {
            Point p = fLocator.locate(fParent);
            p.x -= size().width/2 + fOriginX;
            p.y -= size().height/2 + fOriginY;

            if (p.x != 0 || p.y != 0) {
                willChange();
                basicMoveBy(p.x, p.y);
                changed();
            }
        }
    }

    public void release() {
        super.release();
        if (fParent != null) {
            fParent.removeChild(this);
	}
    }

    /**
     * Creates the current font to be used for new text figures.
     */
    static public Font createCurrentFont() {
        return Fontkit.getFont(fgCurrentFontName, fgCurrentFontStyle, fgCurrentFontSize);
    }

    /**
     * Sets the current font name
     */
    static public void setCurrentFontName(String name) {
        fgCurrentFontName = name;
    }

    /**
     * Sets the current font size.
     */
    static public void setCurrentFontSize(int size) {
        fgCurrentFontSize = size;
    }

    /**
     * Sets the current font style.
     */
    static public void setCurrentFontStyle(int style) {
        fgCurrentFontStyle = style;
    }


    public void figureChanged(FigureChangeEvent e) {
        updateLocation();
    }

    public void figureHandlesChanged(FigureChangeEvent e) {
    }

    public void figureRemoved(FigureChangeEvent e) {
        if (listener() != null)
            listener().figureRequestRemove(new FigureChangeEvent(this));
    }

    public void figureRequestRemove(FigureChangeEvent e) {}
    public void figureInvalidated(FigureChangeEvent e) {}
    public void figureRequestUpdate(FigureChangeEvent e) {}

  public boolean inspect(DrawingView view, boolean alternate) {
     if (!alternate) {
        return super.inspect(view,false);
     } else {
      if (acceptsTyping()) {
	((DrawApplication)view.editor()).doTextEdit(this);
	return true;
      }
      return false;
     }
  }


}
