package de.renew.database;

import java.io.*;
import java.util.*;
import java.rmi.*;
import de.renew.shadow.*;
import de.renew.simulator.*;
import de.renew.remote.*;
import de.renew.util.*;

/**
 * This class contains some static utility methods that allow
 * the caller to reestablish a simulation state.
 */

public class SetupHelper {
  /**
   * This class is totally static. One must not create instances
   * of it.
   */
  private SetupHelper() {}

  /**
   * Load a set of nets from disk in the form of a shadow net system.
   * Then compile the net system to create the semantic net level.
   *
   * @param filename a string that denotes that file to be loaded
   * @exception java.io.IOException if the file access failed
   * @exception java.lang.ClassNotFoundException if the deserialization
   *   failed
   * @exception de.renew.shadow.SyntaxException if the loaded net did not
   *   compile correctly.
   */
  public static void setupNets(String filename) throws IOException,
  ClassNotFoundException, SyntaxException
  {
    // Read the shadow net system from disk.
    FileInputStream stream=new FileInputStream(filename);
    ObjectInput input=new ObjectInputStream(stream);
    ShadowNetSystem netSystem=(ShadowNetSystem)input.readObject();
    // Compile the net system.
    netSystem.compile();
    // The nets have automatically been registered and can be queried by
    // Net.forName(String).
  }

  /**
   * This method restores a set of nets from a database source.
   * For the correct usage of this method it is required
   * that no simulator is currently running. Additionally,
   * a transaction source must not yet have been established,
   * otherwise the database might get corrupted.
   *
   * @param source the database source
   */
  public static void restoreNetInstances(RestoreSource source) {
    // Create all net instances with their initial markings.
    String[] netIDs=source.getAllNetIDs();

    // Create arrays for the reestablished nets.
    int n=netIDs.length;
    Net[] nets=new Net[n];
    NetInstance[] instances=new NetInstance[n];
    NetInstanceMap map=new NetInstanceMap();

    // Create all net instances with their initial markings.
    for (int i=0;i<n;i++) {
      String id=netIDs[i];
      nets[i]=Net.forName(source.getNetName(id));
      instances[i]=nets[i].buildInstance();
      instances[i].setID(id);
      map.put(id,instances[i]);
    }

    // Change to the saved marking.
    for (int i=0;i<n;i++) {
      NetInstance instance=instances[i];

      Enumeration places=nets[i].places();
      while (places.hasMoreElements()) {
	Place place=(Place)places.nextElement();
	PlaceInstance pi=instance.getInstance(place);

	pi.lock.lock();
	// Remove old tokens and discard them.
	pi.extractAllTokens(new Vector(),new Vector());
	// Add new tokens.
	Vector ids=new Vector();
	Vector tokens=new Vector();
	source.fillinAllTokens(pi,map,ids,tokens);
	int num=tokens.size();
	for (int j=0;j<num;j++) {
	  // The time does not matter.
	  pi.insertTokenWithID(tokens.elementAt(j),
			       (String)ids.elementAt(j),0);
	}
	pi.lock.unlock();
      }
    }
  }

  /**
   * This method is the main entry point of this class. It simplifies
   * the setup of a remotely accessible simulation from a database
   * with certain net instances and tokens.
   *
   * It needs an object that provides the current database
   * state, an object that creates database transactions, and an
   * RMI URL where an accessor to an automatically created
   * concurrent simulation run can be registered.
   *
   * Typical use:
   * <code><pre>
   * public static void main(String[] args) {
   *   SetupHelper.setup("/some/where/any.sns",
   *                     new MyRestoreSource("ADatabaseURL"),
   *                     new MyTransactionStrategy("ADatabaseURL"),
   *                     new MyIDFactory("ADatabaseURL"),
   *                     "rmi:aWellKnownName");
   * }
   * </pre></code>
   *
   * @param filename a string that denotes that file to be loaded
   * @param source the object that provides access to the database
   * @param transactionStrategy the object that handles database transactions
   * @param idFactory the object that creates new IDs for net instances
   *   and tokens
   * @param theRMIURL a string that may be used to locate
   *   the simulation, or null if the simulation must not be accessed
   *   remotely
   * @exception java.io.IOException if the file access failed
   * @exception java.lang.ClassNotFoundException if the deserialization
   *   failed
   * @exception de.renew.shadow.SyntaxException if the loaded net did not
   *   compile correctly.
   */
  public static void setup(String filename,
			   RestoreSource source,
			   TransactionStrategy transactionStrategy,
			   IDFactory idFactory,
			   String theRMIURL)
    throws IOException, ClassNotFoundException, SyntaxException
  {
    // Register the ID factory first. It cannot do any harm,
    // although it is probably not needed right now.
    IDSource.setFactory(idFactory);

    // The semantic net layer can now be established.
    setupNets(filename);
	
    // At this point, logging of the net state
    // changes must not be active, because the next method must
    // create net instances and modify them. Such modifications
    // should of course not appear in the database.
    //
    // Moreover, the simulator must not be active yet, or the
    // setup of the simulation state would have to take into account
    // concurrent actions.
    restoreNetInstances(source);

    // We can now enable database logging.
    TransactionSource.setStrategy(transactionStrategy);
    // All subsequent modifications will be recorded in the database.

    // The simulator is now started. We use the concurrent
    // simulator, because database backed application will usually
    // have a need for concurrency.
    Simulator simulator=new ConcurrentSimulator(SilentTracerFactory.INSTANCE);
    simulator.startRun();
	
    if (theRMIURL!=null) {
      // A URL was provided where a simulation accessor can be
      // provided. Let's register an accessor object at the
      // RMI registry.
      SimulatorAccessor accessor=new SimulatorAccessorImpl(simulator);
      Naming.rebind(theRMIURL,accessor);
    }
  }
}







