package de.renew.gui;

import de.renew.simulator.Net;
import de.renew.simulator.NetInstance;

import de.renew.simulator.Place;
import de.renew.simulator.PlaceInstance;
import de.renew.event.PlaceEventProducer;

import de.renew.simulator.Transition;
import de.renew.simulator.TransitionInstance;
import de.renew.event.TransitionEventProducer;

import de.renew.shadow.ShadowLookup;
import de.renew.shadow.ShadowPlace;
import de.renew.shadow.ShadowTransition;

import java.util.Vector;
import java.util.Enumeration;
import java.awt.Menu;
import java.awt.MenuItem;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import CH.ifa.draw.util.CommandMenu;
import CH.ifa.draw.framework.Figure;
import CH.ifa.draw.framework.FigureWithID;
import CH.ifa.draw.framework.FigureEnumeration;

/**
 * Manages Breakpoints and keeps all information together
 * needed by <code>Breakpoint</code> objects.
 * <p>
 * Supports two kinds of breakpoints:<ul>
 * <li>Listener:     Breakpoint objects listening to transition
 *                   or place events, hittable within the one
 *                   net instance to which they are bound or
 *                   within all instances of a net, depending
 *                   on the observed net element.
 *                   These objects are added without modifying
 *                   the compiled net.
 *                   Used in any case.</li>
 * <li>Inscriptions: Breakpoint inscriptions added to the net
 *                   immediately after compilation, which will
 *                   be hit in every instance of the net.
 *                   This technique is no longer in use because
 *                   the listener technique is able to cover
 *                   this case, too.</li>
 * </ul></p>
 *
 * BreakpointManager.java
 * Created: Tue May 23  2000
 * @author Michael Duvigneau
 * @see Breakpoint
 **/
public class BreakpointManager  {

  private CPNSimulation simulation;
  private CPNApplication application;
  private Menu breakpointMenu = null;
  private Vector breakpoints = new Vector();

  /**
   * Keeps the hit messages generated by breakpoints
   * to be shown in a single window.
   * This may be wanted if several breakpoints occur
   * in a very short time span. Instead of a single 
   * breakpoint hit message all messages will be returned.
   **/
  private StringBuffer messageMemory = new StringBuffer();

  public BreakpointManager(CPNSimulation simulation, CPNApplication application) {
    this.simulation = simulation;
    this.application = application;
  }

  /**
   * Number of menu entries which are not breakpoints, needed
   * to calculate the position of breakpoints in the menu.
   * Keep consistent with code in <code>getMenu()</code>!
   * @see #getMenu
   * @see #removeBreakpoint
   **/
  private static final int staticMenuEntries = 4;

  /**
   * Returns the breakpoint menu of this manager.
   * It is created if neccessary.
   **/
  public Menu getMenu() {
    if (breakpointMenu==null) {

      // Take care of the final variable staticMenuEntries some
      // lines above if you change the menu composition.

      CommandMenu menu = application.createCommandMenu("Breakpoints");
      menu.add(createSetMenus("Set BP at selection"));
      menu.add(createClearMenus("Clear BP at selection"));
      menu.add(application.createMenuItem("Clear all BPs in current simulation", 
			      new ActionListener() {
                                public void actionPerformed(ActionEvent e) {
				  deleteAllBreakpoints();
				}
                              }));
      menu.addSeparator();
      breakpointMenu = menu;
    }
    return breakpointMenu;
  }

  /**
   * Returns a menu consisting of two setMenus, one for
   * global breakpoints, one for local ones.
   **/
  protected CommandMenu createSetMenus(String title) {
    CommandMenu menu = application.createCommandMenu(title);
    menu.add(createSetMenu("local", ToggleBreakpointCommand.LOCAL));
    menu.add(createSetMenu("global", ToggleBreakpointCommand.GLOBAL));
    return menu;
  }

  /**
   * Returns a menu with set breakpoint commands respecting
   * the given localGlobal mode, which should be one of
   * the ToggleBreakpointCommand constants LOCAL or GLOBAL.
   * The commands cover all known Breakpoint modes.
   *
   * @see ToggleBreakpointCommand#LOCAL
   * @see ToggleBreakpointCommand#GLOBAL
   **/
  protected CommandMenu createSetMenu(String title, int localGlobal) {
    CommandMenu menu = application.createCommandMenu(title);
    menu.add(new ToggleBreakpointCommand("default (t,p)", 
					 application,
					 this,
					 ToggleBreakpointCommand.ADD,
					 localGlobal));
    menu.add(new ToggleBreakpointCommand("firing starts (t)", 
					 application,
					 this,
					 ToggleBreakpointCommand.ADD,
					 localGlobal,
					 Breakpoint.FIRE));
    menu.add(new ToggleBreakpointCommand("firing completes (t)", 
					 application,
					 this,
					 ToggleBreakpointCommand.ADD,
					 localGlobal,
					 Breakpoint.FIRECOMPLETE));
    menu.add(new ToggleBreakpointCommand("marking changes (p)", 
					 application,
					 this,
					 ToggleBreakpointCommand.ADD,
					 localGlobal,
					 Breakpoint.MARKINGCHANGE));
    menu.add(new ToggleBreakpointCommand("marking changes, ignoring test arcs (p)", 
					 application,
					 this,
					 ToggleBreakpointCommand.ADD,
					 localGlobal,
					 Breakpoint.MARKINGCHANGENOTEST));
    menu.add(new ToggleBreakpointCommand("+ 1 token (p)", 
					 application,
					 this,
					 ToggleBreakpointCommand.ADD,
					 localGlobal,
					 Breakpoint.TOKENADDED));
    menu.add(new ToggleBreakpointCommand("- 1 token (p)", 
					 application,
					 this,
					 ToggleBreakpointCommand.ADD,
					 localGlobal,
					 Breakpoint.TOKENREMOVED));
    menu.add(new ToggleBreakpointCommand("test status changes (p)", 
					 application,
					 this,
					 ToggleBreakpointCommand.ADD,
					 localGlobal,
					 Breakpoint.TOKENTESTCHANGE));
    return menu;
  }

  /**
   * Returns a menu consisting of two clearMenus, one for
   * global breakpoints, one for local ones.
   **/
  protected CommandMenu createClearMenus(String title) {
    CommandMenu menu = application.createCommandMenu(title);
    menu.add(createClearMenu("local", ToggleBreakpointCommand.LOCAL));
    menu.add(createClearMenu("global", ToggleBreakpointCommand.GLOBAL));
    return menu;
  }

  /**
   * Returns a menu with clear breakpoint commands respecting
   * the given localGlobal mode, which should be one of
   * the ToggleBreakpointCommand constants LOCAL or GLOBAL.
   *
   * @see ToggleBreakpointCommand#LOCAL
   * @see ToggleBreakpointCommand#GLOBAL
   **/
  protected CommandMenu createClearMenu(String title, int localGlobal) {
    CommandMenu menu = application.createCommandMenu(title);
    menu.add(new ToggleBreakpointCommand("any type", 
					 application,
					 this,
					 ToggleBreakpointCommand.REMOVE,
					 localGlobal));
    return menu;
  }

  /**
   * Returns a menu with show and clear breakpoint commands
   * concerning the specified breakpoint.
   **/
  protected CommandMenu createBPMenu(final Breakpoint bp) {
    CommandMenu menu = application.createCommandMenu(bp.toString());
    menu.add(application.createMenuItem("show", 
	     new ActionListener() {
		 public void actionPerformed(ActionEvent e) {
		   showBreakpoint(bp);
		 }
	     }));
    menu.add(application.createMenuItem("clear", 
	     new ActionListener() {
		 public void actionPerformed(ActionEvent e) {
		   deleteBreakpoint(bp);
		 }
	     })); 
    return menu;
  }

  /** 
   * Creates one breakpoint at the given transition instance.
   *
   * The created breakpoint is added to the internal list of
   * known breakpoints automatically.
   *
   * @param instance the transition instance to observe
   * @param mode     on which event to react on, given as
   *                 <code>Breakpoint</code> mode.
   *
   * @return null if the mode was not valid,
   *         a breakpoint object otherwise.
   *
   * @see #isValidTransitionMode
   **/
  Breakpoint createTransitionInstanceBreakpoint(TransitionInstance instance,
						int mode) {
    return createTransitionListenerBreakpoint(instance, mode,
					      instance.getNetInstance().getNet());
  }

  /** 
   * Creates one breakpoint at the given transition which
   * will be hit at any instance of that transition.
   *
   * The created breakpoint is added to the internal list of
   * known breakpoints automatically.
   *
   * @param transition the transition to observe
   * @param mode       on which event to react on, given as
   *                   <code>Breakpoint</code> mode.
   * @param net        the net the transition belongs to
   *
   * @return null if the mode was not valid,
   *         a breakpoint object otherwise.
   *
   * @see #isValidTransitionMode
   **/
  Breakpoint createTransitionBreakpoint(Transition transition, 
					int mode, 
					Net net) {
    return createTransitionListenerBreakpoint(transition, mode, net);
  }

  /**
   * The common method body of 
   * @see #createTransitionBreakpoint
   * @see #createTransitionInstanceBreakpoint.
   **/
  private Breakpoint createTransitionListenerBreakpoint(
                     TransitionEventProducer producer, 
		     int mode,
		     Net net) {
    if (! isValidTransitionMode(mode))
      return null;

    TransitionInstanceBreakpoint newBreakpoint
      = new TransitionInstanceBreakpoint(this, producer, mode, net);
    addBreakpoint(newBreakpoint);
    return newBreakpoint;
  }

  /** 
   * Creates one breakpoint at the given place instance.
   *
   * The created breakpoint is added to the internal list of
   * known breakpoints automatically.
   *
   * @param instance the place instance to observe
   * @param mode     on which event to react on, given as
   *                 <code>Breakpoint</code> mode.
   *
   * @return null if the mode was not valid,
   *         a breakpoint object otherwise.
   *
   * @see #isValidPlaceMode
   **/
  public Breakpoint createPlaceInstanceBreakpoint(
                    PlaceInstance instance, int mode) {
    return createPlaceListenerBreakpoint(instance, mode, 
					 instance.getNetInstance().getNet());
  }

  /** 
   * Creates one breakpoint at the given place, which will
   * be hit at any instance of that place.
   *
   * The created breakpoint is added to the internal list of
   * known breakpoints automatically.
   *
   * @param place the place to observe
   * @param mode  on which event to react on, given as
   *              <code>Breakpoint</code> mode.
   * @param net   the net the the place belongs to
   *
   * @return null if the mode was not valid,
   *         a breakpoint object otherwise.
   *
   * @see #isValidPlaceMode
   **/
  public Breakpoint createPlaceBreakpoint(Place place, 
					  int mode, 
					  Net net) {
    return createPlaceListenerBreakpoint(place, mode, net);
  }

  /**
   * The common method body of 
   * @see #createPlaceBreakpoint
   * @see #createPlaceInstanceBreakpoint
   **/
  private Breakpoint createPlaceListenerBreakpoint(
                     PlaceEventProducer producer, 
		     int mode,
		     Net net) {
    if (! isValidPlaceMode(mode))
      return null;

    PlaceInstanceBreakpoint newBreakpoint
      = new PlaceInstanceBreakpoint(this, producer, mode, net);
    addBreakpoint(newBreakpoint);
    return newBreakpoint;
  }

  /** 
   * Creates one global breakpoint at the given transition.
   * The created breakpoint is added to the internal list of
   * known breakpoints automatically.
   * <p>
   * <b>Caution:</b>
   * As there will some inscriptions be removed from the net,
   * the simulation must not run while this method executes!
   * The same condition must be met when the breakpoint is
   * to be deleted.
   * </p>
   *
   * @param transition the transition to observe
   * @param mode       on which events to react on, given as
   *                   <code>Breakpoint</code> modes.
   *                   Valid values are:
   *                   DEFAULT, INSCRIPTION.
   * @param net        the net to which the transition belongs
   *
   * @return <code>null</code> if the mode was not valid,
   *         a breakpoint object otherwise.
   *
   * @see Breakpoint#DEFAULT
   * @see Breakpoint#INSCRIPTION
   *
   * @deprecated 
   * Use <code>createTransitionBreakpoint(transition, 
   * mode, net)</code> instead.
   * @see #createTransitionBreakpoint
   **/
  public GlobalTransitionBreakpoint createGlobalTransitionBreakpoint
                                    (Transition transition, 
				     int mode, 
				     Net net) {
    if ( (mode != Breakpoint.INSCRIPTION)
       & (mode != Breakpoint.DEFAULT) ) 
      return null;

    GlobalTransitionBreakpoint newBreakpoint
      = new GlobalTransitionBreakpoint(this, transition, net);
    addBreakpoint(newBreakpoint);
    return newBreakpoint;
  }

  /**
   * Deletes all breakpoints attached to the given place,
   * transition or place or transition instance.
   * <p>
   * Will not cross the global/local distinction: If a
   * transition is given, breakpoints attached to local
   * instances of that transition will not be deleted.
   * Naturally, this holds also the other way around.
   * </p>
   * @param netElement the place, transition or place or
   *                   transition instance which should
   *                   be freed of breakpoints.
   * @return number of deleted breakpoints
   **/
  int deleteBreakpointsAt(Object netElement) {
    Breakpoint currentBreakpoint;
    int count = 0;
    // Deleting from bottom to top of list to avoid
    // position moves while traversing the list.
    for (int i = breakpoints.size() - 1; i >= 0; i--) {
      currentBreakpoint = (Breakpoint)breakpoints.elementAt(i);
      if (currentBreakpoint.getTaggedElement() == netElement) {
	currentBreakpoint.release();
	removeBreakpoint(i);
	count++;
      }
    }
    return count;
  }

  /** 
   * Clears all currently known breakpoints.
   * This method should be called at least when the
   * current simulation is terminated.
   **/
  public void deleteAllBreakpoints() {
    Breakpoint currentBreakpoint;
    // Deleting from bottom to top of list to avoid
    // position moves while traversing the list.
    for (int i = breakpoints.size() - 1; i >= 0; i--) {
      currentBreakpoint = (Breakpoint)breakpoints.elementAt(i);
      currentBreakpoint.release();
      removeBreakpoint(i);
    }
  }

  /**
   * Clears the specified breakpoint. 
   * Releases its resources and removes it from
   * the list of known breakpoints.
   **/
  public void deleteBreakpoint(Breakpoint bp) { 
    Breakpoint currentBreakpoint;
    for (int i = breakpoints.size() - 1; i >= 0; i--) {
      currentBreakpoint = (Breakpoint)breakpoints.elementAt(i);
      if (currentBreakpoint == bp) {
	removeBreakpoint(i);
      }
    }
    bp.release();
  }

  /**
   * Attaches global breakpoints to all places or transitions
   * with breakpoint attribute at the corresponding graphical
   * figure.
   * Traverses the drawings of all nets included in the
   * given shadow lookup, if they are opened.
   * <p>
   * There is no check if any of these breakpoints already exist.
   * So this method should be called only once per simulation.
   * </p>
   **/
  protected void addPresetBreakpoints(ShadowLookup lookup) {
    Enumeration netNames = lookup.allNetNames();
    String currentNetName;
    CPNDrawing currentDrawing;
    FigureEnumeration figures;
    Figure currentFigure;
    Integer currentMode;
    TransitionFigure currentTransFig;
    Transition currentTransition;
    PlaceFigure currentPlaceFig;
    Place currentPlace;

    while (netNames.hasMoreElements()) {
      currentNetName = (String)netNames.nextElement();
      currentDrawing = simulation.getDrawing(currentNetName);
      figures = currentDrawing.figures();
      while (figures.hasMoreElements()) {
	currentFigure = figures.nextFigure();
	currentMode = (Integer)currentFigure.getAttribute(Breakpoint.ATTRIBUTENAME);
	if (currentMode != null) {
	  if (currentFigure instanceof TransitionFigure) {
	    currentTransFig = (TransitionFigure)currentFigure;
	    currentTransition = lookup.get((ShadowTransition)currentTransFig.getShadow());
System.out.println("Creating Breapoint: "+
	    createTransitionBreakpoint(currentTransition, 
				       currentMode.intValue(), 
				       lookup.getNet(currentNetName)));
	    
	  } else if (currentFigure instanceof PlaceFigure) {
	    currentPlaceFig = (PlaceFigure)currentFigure;
	    currentPlace = lookup.get((ShadowPlace)currentPlaceFig.getShadow());
System.out.println("Creating Breapoint: "+
	    createPlaceBreakpoint(currentPlace, 
				  currentMode.intValue(), 
				  lookup.getNet(currentNetName)));
	  }
	}
      }
    }
  }

  /**
   * Appends the given breakpoint to the end of the list
   * of known breakpoints. Keeps the menu consistent.
   **/
  private void addBreakpoint(Breakpoint newBreakpoint) {
    breakpoints.addElement(newBreakpoint);
    breakpointMenu.add(createBPMenu(newBreakpoint));
  }


  /**
   * Removes the breakpoint at the given position from the list
   * of known breakpoints. Keeps the menu consistent.
   **/
  private void removeBreakpoint(int position) {
    breakpoints.removeElementAt(position);
    breakpointMenu.remove(staticMenuEntries + position);
  }

  /**
   * Stops the running simulation (which is what breakpoints 
   * should do) by using a concurrent thread.
   * <p>
   * The extra thread is needed to avoid deadlock situations
   * within <code>de.renew.simulator.Simulator.stopRun()</code>
   * implementations. 
   * The deadlock occurs because the <code>stopRun()</code>
   * methods try to ensure that the simulation has stopped,
   * but the execution of this breakpoint is part of the
   * running simulation - and it would stop only after the
   * <code>stopRun()</code> waiting condition is met...
   * </p><p>
   * <b>Drawback</b>: This method does not guarantee that the 
   * simulation is stopped after the method has finished.
   * </p>
   **/
  protected void stopSimulation() {
System.out.println("BM creates StopThread.");
    new Thread() {
      public void run() {
System.out.println("StopThread runs.");
	simulation.simulationStop();
System.out.println("StopThread is done.");
      }
    }.start();
System.out.println("BM: StopThread created.");
  }

  /**
   * Shows a message dialog telling that the given breakpoint
   * was hit and selects its associated net element figure.
   * Used by breakpoints when they occur.
   *
   * @param breakpoint the breakpoint which has occured.
   **/
  protected void showHitBreakpoint(Breakpoint breakpoint) {
    application.showStatus("A breakpoint was hit: "+breakpoint+".");
    FigureException e = locateBreakpoint(breakpoint, true);
    if (e != null) application.processFigureException(e, true);
  } 

  /**
   * Selects the associated net element figure of the given
   * Breakpoint.
   * Used by breakpoint menu entries of kind "Show breakpoint".
   *
   * @param breakpoint the breakpoint to show.
   **/
  protected void showBreakpoint(Breakpoint breakpoint) {
    application.showStatus("Selecting "+breakpoint+".");
    FigureException e = locateBreakpoint(breakpoint, false);
    if (e != null) application.selectOffendingElements(e);
  }

  /**
   * Empties the memory of generated beakpoint hit messages
   * so they won't be shown again with the next hit breakpoint.
   **/
  public void clearLog() {
    messageMemory = new StringBuffer();
  }

  /** 
   * Looks up the associated net element figure of the given
   * breakpoint and returns it in a <code>FigureException</code>.
   * The hit element of the breakpoint will be used, if <code>hit
   * </code>is <code>true</code>, else the tagged element will
   * be looked up.
   * <p>
   * The FigureException will also include a generated hit message,
   * if requested.
   * If the mapping to a Figure was not possible, the messages
   * are printed to System.err.
   * </p>
   *
   * @param breakpoint the breakpoint to look up.
   * @param hit        if true, an hit message is generated.
   * @return           the figure exception containing the message,
   *                   the element figure and the drawing;
   *                   <code>null</code> if no figure was found.
   *
   * @see #showHitBreakpoint
   * @see #showBreakpoint
   **/
  private FigureException locateBreakpoint (Breakpoint breakpoint, 
					    boolean hit) {
    String message = "";
    String allMessages;
    Net net = null;
    NetInstance netInstance = null;
    int id = FigureWithID.NOID;
    Object netElement = hit?breakpoint.getHitElement()
                           :breakpoint.getTaggedElement();

    if (netElement instanceof PlaceInstance) {
      PlaceInstance placeInst = (PlaceInstance)netElement;
      id = placeInst.getPlace().getID();
      netInstance = placeInst.getNetInstance();
      net = netInstance.getNet();
      if (hit) message = "Hit "+breakpoint+".";
    } else if (netElement instanceof TransitionInstance) {
      TransitionInstance transitionInst = (TransitionInstance)netElement;
      id = transitionInst.getTransition().getID();
      netInstance = transitionInst.getNetInstance();
      net = netInstance.getNet();
      if (hit) message = "Hit "+breakpoint+".";
    } else if (netElement instanceof Transition) {
      Transition transition = (Transition)netElement;
      id = transition.getID();
      net = breakpoint.getTaggedNet();
      // No hit message as static transitions cannot fire.
    } else if (netElement instanceof Place) {
      Place place = (Place)netElement;
      id = place.getID();
      net = breakpoint.getTaggedNet();
      // No hit message as static places cannot change anything.
    } else {
      System.err.println("Cannot determine the location of breakpoint at "+netElement+"!?");
      return null;
    }

    if (hit) {
      allMessages = messageMemory.toString() + "\n" + message;
      messageMemory.append("(");
      messageMemory.append(message);
      messageMemory.append(")\n");
    } else {
      allMessages = message;
    }

    String netName = net.getName();
    CPNDrawing drawing = simulation.getDrawing(netName);
    if (drawing == null) {
      message = message + "\nSorry, cannot show the location of breakpoint at "
                        + netElement+" because net "+netName+" is not loaded.";
      application.showStatus(message);
      System.err.println(message);
      return null;
    } else if (netInstance == null) {
      FigureWithID figure = drawing.getFigureWithID(id);
      return new FigureException("Renew: Breakpoint",
				 allMessages, 
				 drawing,
				 figure);
    } else {
      application.openInstanceDrawing(netInstance);
      CPNInstanceDrawing instanceDrawing =
	CPNInstanceDrawing.getInstanceDrawing(application.getMode(),netInstance);
      FigureWithID figure = drawing.getFigureWithID(id);
      InstanceFigure instanceFigure = instanceDrawing.getInstanceFigure(figure);
      return new FigureException("Renew: Breakpoint",
				 allMessages, 
				 instanceDrawing,
				 instanceFigure);
    }
  }

  /**
   * Tells wether the given mode is valid for transition
   * breakpoints.
   * <p>
   * Valid values are: DEFAULT, FIRE, FIRECOMPLETE.
   * </p>
   *
   * @return <code>true</code>, if the given breakpoint mode
   *                            is applicable to transitions
   *                            or transition instances. <br>
   *         <code>false</code>, otherwise.
   *
   * @see Breakpoint#DEFAULT
   * @see Breakpoint#FIRE
   * @see Breakpoint#FIRECOMPLETE
   **/
  public static boolean isValidTransitionMode(int mode) {
    return (mode == Breakpoint.FIRE)
      ||   (mode == Breakpoint.FIRECOMPLETE)
      ||   (mode == Breakpoint.DEFAULT);
  }

  /**
   * Tells wether the given mode is valid for place
   * breakpoints.
   * <p>
   * Valid values are: DEFAULT, MARKINGCHANGE.
   * </p>
   *
   * @return <code>true</code>, if the given breakpoint mode
   *                            is applicable to places or
   *                            place instances. <br>
   *         <code>false</code>, otherwise.
   *
   * @see Breakpoint#DEFAULT
   * @see Breakpoint#MARKINGCHANGE
   **/
  public static boolean isValidPlaceMode(int mode) {
    return (mode == Breakpoint.MARKINGCHANGE)
      ||   (mode == Breakpoint.MARKINGCHANGENOTEST)
      ||   (mode == Breakpoint.TOKENADDED)
      ||   (mode == Breakpoint.TOKENREMOVED)
      ||   (mode == Breakpoint.TOKENTESTCHANGE)
      ||   (mode == Breakpoint.DEFAULT);
  }
}
