package de.renew.gui;

import java.util.*;
import java.awt.*;
import CH.ifa.draw.standard.*;
import CH.ifa.draw.framework.*;
import de.renew.simulator.*;

import java.io.ObjectOutput;
import java.io.ObjectInput;
import java.io.IOException;
import java.io.StreamCorruptedException;
import de.renew.util.RenewObjectOutputStream;
import de.renew.util.RenewObjectInputStream;
import de.renew.util.StringUtil;

public class CPNInstanceDrawing extends StandardDrawing
                                implements DrawingChangeListener,
					   DrawingContext {

  public static String ID_MACRO="$ID";

  private static Hashtable drawingsByInstance = new Hashtable();

  private NetInstance netInstance;
  private CPNDrawing cpnDrawing;
  private RenewMode mode;
  private Hashtable instanceLookup;

  private CPNInstanceDrawing(RenewMode mode, 
			     NetInstance netInstance, 
			     CPNDrawing drawing)
  {
    this.mode = mode;
    this.cpnDrawing = drawing;
    connect(netInstance);
    cpnDrawing.addDrawingChangeListener(this);
  }

  RenewMode getMode() {
      return mode;
  }

  /** Get the NetInstance associated with this drawing. */
  public NetInstance getNetInstance() {
      return netInstance;
  }
    
  public Rectangle displayBox() {
    return super.displayBox().union(cpnDrawing.displayBox());
  }

  /**
   * Returns the instance drawing associated to the given net
   * instance, creates the drawing if neccessary.
   *
   * @param mode needed to look up the responsible simulation
   * @param netInstance to look up the drawing for
   * @return CPNInstanceDrawing which displays the given netInstance.
   *         If there is currently no such drawing, one is created.
   *         If there is not even any CPNDrawing matching the net
   *         known to the active simulation, returns 
   *         <code>null</code>.
   **/
  static CPNInstanceDrawing getInstanceDrawing(RenewMode mode,
					       NetInstance netInstance)
  {
    // forbid concurrent searches for instance drawings.
    synchronized (drawingsByInstance) {
      // find instance drawing:
      if (drawingsByInstance.containsKey(netInstance))
	return (CPNInstanceDrawing)drawingsByInstance.get(netInstance);
      else {
	// find drawing to base the instance drawing on
	CPNSimulation sim = mode.getSimulation();
	CPNDrawing drawing = sim.getDrawing(netInstance.getNet().getName());
	if (drawing != null) 
	  return new CPNInstanceDrawing(mode, netInstance, drawing);
	else 
	  System.err.println("To display a net instance, an open"
			     +" drawing of the instantiated net is"
			     +" needed. No such drawing found for"
			     +" net \""+netInstance.getNet().getName()
			     +"\".");
	  return null;
      }
    }
  }

  /** 
   * Returns all instance drawings based on the specified CPNDrawing.
   * Returns <code>null</code>, if no such drawing exists.
   **/
  public static Enumeration getDependentInstanceDrawings(CPNDrawing baseDrawing) {
    Enumeration allDrawings = drawingsByInstance.elements();
    Vector dependentDrawings = new Vector();
    while (allDrawings.hasMoreElements()) {
      CPNInstanceDrawing oneDrawing = (CPNInstanceDrawing)allDrawings.nextElement();
      if (oneDrawing.cpnDrawing == baseDrawing)
	dependentDrawings.addElement(oneDrawing);
    }

    if (dependentDrawings.size() > 0)
      return dependentDrawings.elements();
    else
      return null;
  }
  
  public void draw(Graphics g) {
      //g.setColor(Color.yellow.brighter());
      //Rectangle b=displayBox();
      //g.fillRect(b.x,b.y,b.width,b.height);
    cpnDrawing.draw(g,this);
    super.draw(g);
  }

  public InstanceFigure getInstanceFigure(Figure fig) {
      return (InstanceFigure)instanceLookup.get(fig);
  }

  public boolean isHighlighted(Figure fig) {
      FigureWithHighlight node=cpnDrawing.getFigureForHighlight(fig);
      if (node != null)
	  fig = node;
      InstanceFigure instFig=getInstanceFigure(fig);
      if (instFig!=null) {
	  return instFig.isHighlighted();
      }
      return false;
  }

  public boolean isVisible(Figure fig) {
      return !(((fig instanceof CPNTextFigure
		 && ((CPNTextFigure)fig).getType() == CPNTextFigure.INSCRIPTION
		 && ((CPNTextFigure)fig).parent() instanceof PlaceFigure))
	       || fig == cpnDrawing.getIconFigure());
  }

  public String expandMacro(String text) {
      return expandMacro(text,netInstance);
  }

  public static String expandMacro(String text, NetInstance instance) {
      return StringUtil.replace(text,ID_MACRO,instance.getID());
  }

  /**
   * Creates instance figures in this instance drawing which
   * reflect the net elements of the given net instance.
   * Used by the (private) constructor to build an instance
   * drawing.
   * <p>
   * Also... <ul>
   * <li> builds the instance lookup table
   *      (see <code>getInstanceFigure()</code>) </li>
   * <li> registers this instance drawing in the
   *      (private) drawingsByInstance map
   *      (see <code>getInstanceDrawing()</code>) </li>
   * </ul></p>
   * May be more, but that is what I could see in the code.
   *
   * @param netInstance the net instance to be displayed
   * @see #getInstanceFigure
   * @see #getInstanceDrawing
   **/
  public void connect(NetInstance netInstance) {

    if (netInstance != null) {
      if (this.netInstance != null)
	drawingsByInstance.remove(this.netInstance);
      this.netInstance = netInstance;
      drawingsByInstance.put(netInstance,this);
      setName(netInstance.toString());
      instanceLookup = new Hashtable();
      // FigureLookup figureLookup = mode.getSimulation().getFigureLookup();
      // The global FigureLookup is deprecated since we have the ID mapping.
      // Now each net instance drawing can build its own lookup...
      Hashtable figureLookup = buildFigureLookup(netInstance);
      Enumeration figures = figureLookup.keys();
      while (figures.hasMoreElements()) {
	Figure figure = (Figure)figures.nextElement();
	if (figure instanceof PlaceFigure) {
	  PlaceFigure placeFig = (PlaceFigure)figure;
	  Place place=
	    (Place)figureLookup.get(placeFig.getSemanticPlaceFigure());
	  if (place!=null) {
	    PlaceInstance placeInst = netInstance.getInstance(place);
	    Figure pif=new PlaceInstanceFigure(this,placeFig,placeInst);
	    instanceLookup.put(placeFig,pif);
	    add(pif);
	    sendToBack(pif);
	  }
	} else if (figure instanceof TransitionFigure) {
	  TransitionFigure transitionFig = (TransitionFigure)figure;
	  Transition transition=(Transition)figureLookup.get(transitionFig);
	  if (transition!=null) {
	    TransitionInstance transitionInst = netInstance.getInstance(transition);
            Figure tif=new TransitionInstanceFigure(this,transitionFig,transitionInst);
	    instanceLookup.put(transitionFig,tif);
	    add(tif);
	  }
	}
      }
    }
  }


  /**
   * Tries to map graphical figures of an open CPNDrawing
   * to net elements of the given NetInstance.
   *
   * @param instance  NetInstance to try to map
   * @return          Mapping from figures to net elements
   *                  (which is empty if the mapping was
   *                  unsuccessfull)
   **/
  protected Hashtable buildFigureLookup(NetInstance netInstance) {

    Hashtable idMapping = null;
    Hashtable figureMapping = null;

    Net net = netInstance.getNet();
    String netName = net.getName();
    CPNDrawing netDrawing = mode.getSimulation().getDrawing(netName);
    if (netDrawing == null) {
      System.err.println("No open drawing found matching the "
			 +" compiled net \""+netName+"\"");
    } else {
      //System.out.println("Mapping graphical to compiled "
      //                  +"elements for net \""+netName
      //                  +"\"...");
      Enumeration  netElements;
      Place        currentPlace;
      Transition   currentTransition;
      int          currentID;
      boolean warning1Found, warning2Found, warning3Found;
      boolean errorFound;
      int returnStatus;

      // Proceeding: First build a mapping from IDs to
      // net elements, then try to map net figures
      // to these elements using their ID.

      idMapping = new Hashtable();
      figureMapping = new Hashtable();

      // Fill in all places into the ID mapping
      netElements = net.places();
      while (netElements.hasMoreElements()) {
	currentPlace = (Place)netElements.nextElement();
	currentID = currentPlace.getID();
	idMapping.put(new Integer(currentID), currentPlace);
      }

      // Fill in all transitions into the ID mapping
      netElements = net.transitions();
      while (netElements.hasMoreElements()) {
	currentTransition = (Transition)netElements.nextElement();
	currentID = currentTransition.getID();
	idMapping.put(new Integer(currentID), currentTransition);
      }

      // Try to map figures to net elements by ID
      // There are 4 problem categories:
      //
      // 1 Missing figure for a net element
      //   (e.g. the user added elements to the drawing)
      // 2 Missing net element for a figure
      //   (e.g. the user removed elements from the drawing)
      // 3 Different types of elements with the same ID
      //   (e.g. the user removed and added elements)
      // 4 Same element type, but modified attributes:
      //   Such errors are neither recognized nor handled.
      //
      // If one of the problems 1-3 occurs, it is
      // impossible to map the offending net element.
      // But as other elements may still be mapable,
      // a warning message seems to be appropriate.
      //
      // There is currently no problem resulting in
      // an fatal error.
      errorFound = false;
      warning1Found = false;
      warning2Found = false;
      warning3Found = false;
      returnStatus = 0;

      // Check for problem type 1:
      // Get all IDs of the ID mapping and check if
      // corresponding figures exist.
      // This check is done in an extra pass because it needs
      // the inverse relation.
      netElements = idMapping.keys();
      while (!errorFound && netElements.hasMoreElements()) {
	currentID = ((Integer)netElements.nextElement()).intValue();
	if (netDrawing.getFigureWithID(currentID) == null) {
	  //System.err.println("Missing figure: "+currentID);		
	  warning1Found = true;
	}
      }

      // Finally, do the mapping.
      // This main step is outsourced because of the recursion
      // which is necessary when CompositeFigures are encountered.
      returnStatus = addToFigureLookup(netDrawing,
				      figureMapping,
				      idMapping);
      errorFound    = errorFound    | ((returnStatus & 1) > 0);
      warning1Found = warning1Found | ((returnStatus & 2) > 0);
      warning2Found = warning2Found | ((returnStatus & 4) > 0);
      warning3Found = warning3Found | ((returnStatus & 8) > 0);

      // Print the appropriate warning messages
      if (warning1Found) {
	System.err.println("Warning: Drawing \""+netName+"\""
			   +" does not contain one or more net"
			   +" elements found in the compiled net.");
	System.err.println("Some logical elements will not be"
			   +" displayed.");
      }
      if (warning2Found) {
	System.err.println("Warning: Drawing \""+netName+"\""
			   +" contains graphical net elements"
			   +" not found in the compiled net.");
	System.err.println("Some graphical elements will not be"
			   +" functional.");
      }
      if (warning3Found) {
	System.err.println("Warning: Drawing \""+netName+"\""
			   +" contains graphical elements which map"
			   +" to compiled net elements of incompatible"
			   +" type.");
	System.err.println("Some graphical elements will not be"
			   +" functional.");
      } 

      // If some fatal problem was encountered,
      // deliver an empty mapping instead of an invalid
      // one. 
      // But there is no such error detected at this time.
      if (errorFound) {
	System.err.println("Error: Drawing \""+netName+"\""
			   +" not mappable due to a \"THIS"
			   +"-CANNOT-HAPPEN-ERROR\"!?");
	figureMapping.clear();
      }
    }

    return figureMapping;
  }

  /**
   * This method should not be called directly, it is
   * used by <code>buildFigureLookup</code>.
   * @param  container      CompositeFigure to traverse,
   *                        this method checks all its 
   *                        figures.
   * @param  figureMapping  Hashtable to extend with
   *                        additional pairs of the type
   *                        (figure, net element).
   * @param  idMapping      Hashtable used to look up
   *                        net elements by their id.
   * @return status -       Internal bitcode reflecting
   *                        the encountered problems.
   **/
  private int addToFigureLookup(final CompositeFigure container,
				Hashtable             figureMapping,
				final Hashtable       idMapping) {

    Enumeration  figures;
    Figure       aFigure;
    FigureWithID currentFigure;
    Place        currentPlace;
    Transition   currentTransition;

    // Description of the warning types is
    // located in buildFigureLookup
    boolean warning1Found = false;
    boolean warning2Found = false;
    boolean warning3Found = false;
    boolean errorFound    = false;
    int    returnStatus  = 0;

    // Iterate through all figures and look up the
    // corresponding (with same ID) net elements.
    figures = container.figures();
    while (!errorFound && figures.hasMoreElements()) {
      // If there are figures without ID support,
      // they are not relevant, skip them.
      // But CompositeFigures contain other figures,
      // so recurse.
      aFigure = (Figure)figures.nextElement();
      if (aFigure instanceof FigureWithID) {
	currentFigure = (FigureWithID)aFigure;
      } else if (aFigure instanceof CompositeFigure) {
	returnStatus = returnStatus
	             | addToFigureLookup((CompositeFigure)aFigure,
					 figureMapping,
					 idMapping);
	errorFound = errorFound | ((returnStatus & 1) > 0);
	currentFigure = null;
      } else {
	currentFigure = null;
      }

      // If we have a figure with ID, we can go deeper...
      if (currentFigure != null) {
	// Check only places and transitions as only their
	// IDs are in the idMapping.
	// As there is no common superclass supporting IDs,
	// the code has to be written twice...
	if (currentFigure instanceof PlaceFigure) {
	  int currentID =
	    ((PlaceFigure)currentFigure).getSemanticPlaceFigure().getID();
	  try {
	    currentPlace = (Place)idMapping.get(new Integer(currentID));
	    if (currentPlace != null) {
	      figureMapping.put(currentFigure, currentPlace);
	    } else {
	      //System.err.println("Missing net element: "+currentID);		
	      warning2Found = true;
	    }
	  } catch (ClassCastException e) {
	    //System.err.println("Mismatch: "+currentID);		
	    warning3Found = true;
	  }
	} else if (currentFigure instanceof TransitionFigure) {
	  int currentID = currentFigure.getID();
	  try {
	    currentTransition = (Transition)idMapping.get(new Integer(currentID));
	    if (currentTransition != null) {
	      figureMapping.put(currentFigure, currentTransition);
	    } else {
	      //System.err.println("Missing net element: "+currentID);		
	      warning2Found = true;
	    }
	  } catch (ClassCastException e) {
	    //System.err.println("Mismatch: "+currentID);		
	    warning3Found = true;
	  }
	}
      }
    }

    if (errorFound)    returnStatus = returnStatus | 1;
    if (warning1Found) returnStatus = returnStatus | 2;
    if (warning2Found) returnStatus = returnStatus | 4;
    if (warning3Found) returnStatus = returnStatus | 8;
    return returnStatus;
  }

  public void release() {
    cpnDrawing.removeDrawingChangeListener(this);
    if (netInstance != null) {
      drawingsByInstance.remove(netInstance);
      //System.err.println("Instance Drawing removed from list.");
      instanceLookup=null;
      netInstance = null;
    }
    super.release();
  }


  /**
   *  Sent when an area is invalid
   */
  public void drawingInvalidated(DrawingChangeEvent e) {
    // System.out.println("Source invalidated: "+e.getInvalidatedRectangle());
    DrawingChangeEvent myEvent = new DrawingChangeEvent(this,
      e.getInvalidatedRectangle());
    Enumeration listeners = drawingChangeListeners();
    while (listeners.hasMoreElements()) {
      DrawingChangeListener listener = (DrawingChangeListener)listeners.nextElement();
      listener.drawingInvalidated(myEvent);
      listener.drawingRequestUpdate(myEvent);
    }
  }

  /**
   *  Sent when the drawing wants to be refreshed
   */
  public void drawingRequestUpdate(DrawingChangeEvent e) {
    //System.out.println("Source requests update: "+e.getInvalidatedRectangle());
  }


   /**
     * Returns whether drawing has been modified since last save.
     */
    public boolean isModified() {
       return false;  // an InstanceDrawing should not be saved!
    }

  /**
   * Writes all currently known <code>NetInstances</code>s
   * to the given stream. The written information is <b>not</b>
   * guaranteed to describe the complete current simulation
   * state!
   * A <code>NetInstances</code> is known, if it is displayed
   * by a <code>CPNInstanceDrawing</code>.
   * <p>
   * If the given <code>ObjectOutput</code> is a <code><b>
   * de.renew.util.RenewObjectOutputStream</b></code>, its
   * feature of cutting down the recursion depth by delaying
   * the serialization of some fields will be used.
   * </p><p>
   * <b>Side effect:</b> Any open <code>BindingSelectionFrame</code>
   * will be closed on execution of this method!
   * <br></br>
   * <b>Caution:</b> In order to get consistent data written
   * to the stream, you have to ensure that there are no
   * concurrent modifications of the simulation state.
   * This method is not able to lock the simulation.
   * </p>
   *
   * Added Feb 29 2000  Michael Duvigneau
   *
   * @param output target stream (see note about RenewObjectOutputStream)
   * @see de.renew.util.RenewObjectOutputStream
   **/
  public static void saveAllInstances(ObjectOutput output) 
         throws IOException {

    Enumeration saveList;
    RenewObjectOutputStream rOut = null;
    if (output instanceof RenewObjectOutputStream)
      rOut = (RenewObjectOutputStream)output;

    if (rOut != null) rOut.beginDomain(CPNInstanceDrawing.class);

    // Close the BindingSelectionFrame (otherwise it would
    // be stored with its transition).
    BindingSelectionFrame.close();

    // Save all currently known instances.
    output.writeInt(drawingsByInstance.size());
    saveList=drawingsByInstance.keys();
    while (saveList.hasMoreElements()) {
      output.writeObject(saveList.nextElement());
    }

    // If a RenewObjectOutputStream is used, write
    // all delayed fields NOW. 
    if (rOut != null) rOut.writeDelayedObjects();

    if (rOut != null) rOut.endDomain(CPNInstanceDrawing.class);
  }

  /**
   * Reads NetInstances stored by <code>saveAllInstances()
   * </code>.
   * <p>
   * This method does neither automatically add the new<code>
   * NetInstance</code>s to the list of known instances nor
   * open a CPNInstanceDrawing for them.
   * The reason is that first the active simulation has to
   * rebuild its FigureLookup.
   * It is then left up to the caller to request the
   * appropriate <code>CPNInstanceDrawing</code>s via
   * <code>getInstanceDrawing()</code>.
   * </p><p>
   * If the given <code>ObjectInput</code> is a <code>
   * de.renew.util.RenewObjectInputStream</code>, the
   * neccessary steps to cover delayed serialization will
   * be made.
   * </p><p>
   * The ObjectInputStream will be read using
   * <code>de.renew.util.ClassSource</code> to provide
   * its ability of reloading all user defined classes.
   * </p>
   *
   * Added Feb 29 2000  Michael Duvigneau
   *
   * @return Enumeration of NetInstances
   * @param input source stream (see note about RenewObjectInputStream)
   * @see #saveAllInstances
   * @see de.renew.util.ClassSource
   * @see de.renew.util.RenewObjectInputStream
   **/
  public static Enumeration loadInstances(ObjectInput input) 
         throws IOException, ClassNotFoundException {

    // Read all stored NetInstances.
    int count = input.readInt();
    Vector newInstances = new Vector(count);
    try {
      for (int i = 0; i < count; i++) {
	newInstances.addElement(
	  (NetInstance)de.renew.util.ClassSource.readObject(input));
      }      
    } catch (ClassCastException e) {
      throw new StreamCorruptedException (
		"Object other than NetInstance found "
		+"when looking for net instances: "
		+e);
    }
    
    // If a RenewObjectInputStream is used, read
    // all delayed fields NOW.
    if (input instanceof RenewObjectInputStream)
      ((RenewObjectInputStream)input).readDelayedObjects();

    // Return the list of read NetInstances.
    return newInstances.elements();
  }
}
