package de.renew.remote;

import java.rmi.*;
import java.rmi.server.*;

import collections.CollectionEnumeration;
import collections.CorruptedEnumerationException;

import de.renew.simulator.Net;
import de.renew.simulator.Place;
import de.renew.simulator.Transition;

/**
 * This class implements the <code>NetAccessor</code> interface
 * and nothing more.
 * <p>
 * </p>
 * NetAccessorImpl.java
 * Created: Thu Jul 13  2000
 * @author Michael Duvigneau
 */

public class NetAccessorImpl extends UnicastRemoteObject
                             implements NetAccessor  {

  /**
   * The net that is accessable via this net accessor.
   */
  private final Net net;

  /**
   * Creates a new net accessor for the given net.
   *
   * @param net the net to access
   * @exception java.rmi.RemoteException if an RMI failure occured.
   */  
  public NetAccessorImpl(Net net) throws RemoteException {
    super();
    this.net = net;
  }

  /* This method is specified by the NetAccessor interface. */
  public String getName() throws RemoteException {
    return net.getName();
  }

  /* This method is specified by the Accessor interface. */
  public String asString() throws RemoteException {
    return net.toString();
  }
  
  // -------------------------------------------- Place handling ----

  /* This method is specified by the NetAccessor interface. */
  public int[] getPlaceIDs() throws RemoteException {

    boolean corrupted;
    CollectionEnumeration placeEnum;
    int[] placeIDs = null;

    // Corruption may occur if the net's set of places changes
    // while copying data into the ID array.
    // As this situation should be very rare, the copying of
    // data is in a loop which will then be restarted.
    // Synchronization is not needed this way.
    // It is important that the first action is getting the
    // enumeration to cover the whole computation with its
    // corruption detection.

    do {
      placeEnum = net.places();
      corrupted = false;
      placeIDs = new int[net.placeCount()];
      try {
	for (int index = 0; index < placeIDs.length; index++)
	  placeIDs[index] = ((Place)placeEnum.nextElement()).getID();
      } catch (CorruptedEnumerationException e) {
	System.err.println("NetAccessor("+net+"): place enumeration corrupted, restarting."); 
	corrupted = true;
      }
    } while (corrupted);

    return placeIDs;
  }

  /**
   * Return a place accessor for the place with the given
   * ID. Return <code>null</code>, if no such place exists.
   *
   * This method is specified by the NetAccessor interface.
   *
   * <p>
   * An additional remark to the validity of the given id:
   * It is not checked if the set of places of the accessed
   * net has changed since the last call to <code>getPlaceIDs()
   * </code>.
   * </p>
   *
   * @see NetAccessor#getPlace
   */
  public PlaceAccessor getPlace(int id) throws RemoteException {
    Place result = net.getPlaceWithID(id);
    if (result==null) return null;
    return new PlaceAccessorImpl(result);
  }

  // --------------------------------------- Transition handling ----

  /* This method is specified by the NetAccessor interface. */
  public int[] getTransitionIDs() throws RemoteException {

    boolean corrupted;
    CollectionEnumeration transitionEnum;
    int[] transitionIDs = null;

    // Corruption may occur if the net's set of transitions changes
    // while copying data into the ID array.
    // As this situation should be very rare, the copying of
    // data is in a loop which will then be restarted.
    // Synchronization is not needed this way.
    // It is important that the first action is getting the
    // enumeration to cover the whole computation with its
    // corruption detection.

    do {
      transitionEnum = net.transitions();
      corrupted = false;
      transitionIDs = new int[net.transitionCount()];
      try {
	for (int index = 0; index < transitionIDs.length; index++)
	  transitionIDs[index] = ((Transition)transitionEnum.nextElement()).getID();
      } catch (CorruptedEnumerationException e) {
	System.err.println("NetAccessor("+net+"): transition enumeration corrupted, restarting."); 
	corrupted = true;
      }
    } while (corrupted);

    return transitionIDs;
  }

  /**
   * Return a transition accessor for the transition with the given
   * ID. Return <code>null</code>, if no such transition exists.
   *
   * This method is specified by the NetAccessor interface.
   *
   * <p>
   * An additional remark to the validity of the given id:
   * It is not checked if the set of transitions of the accessed
   * net has changed since the last call to <code>getTransitionIDs()
   * </code>.
   * </p>
   *
   * @see NetAccessor#getTransition
   */
  public TransitionAccessor getTransition(int id) throws RemoteException {
    Transition result = net.getTransitionWithID(id);
    if (result==null) return null;
    return new TransitionAccessorImpl(result);
  }
 
}
