package de.renew.remote;

import java.rmi.*;

/**
 * An object accessor allows to view some attributes
 * of its associated object.
 *
 * The associated object is not accessible directly but the
 * accessor provides String descriptions of the object as a
 * whole and its fields. This is somewhat like but less
 * powerful than what the java reflection api provides.
 *
 * <p>
 * This is a remote interface that allows an implementation
 * that can be used via RMI.
 * The intention is to keep the associated object on the
 * server and transmit only its description as strings.
 * </p>
 *
 * @author Olaf Kummer
 */
public interface ObjectAccessor extends Accessor {

  /**
   * Return the number of public fields of the associated object.
   * Always zero for primitive values and null references.
   *
   * @return the number of public fields
   * @exception java.rmi.RemoteException if an RMI failure occured.
   */
  public int getFieldCount() throws RemoteException;

  /**
   * Return the names of all fields of the associated object.
   * Always an empty array for primitive values and null references.
   *
   * @return an array of field names
   * @exception java.rmi.RemoteException if an RMI failure occured.
   */
  public String[] getFieldNames() throws RemoteException;

  /**
   * Return the values of all fields of the associated
   * object as a string representation.
   * This produces the same result as getting all fields one 
   * by one via the <code>getField</code> method and putting
   * their <code>asString</code> results into an array.
   *
   * Returns always an empty array for primitive values and
   * null references.
   *
   * @return an array of field value string descriptions
   * @exception java.rmi.RemoteException if an RMI failure occured.
   */
  public String[] getFieldValues() throws RemoteException;

  /**
   * Return an object accessor for the value of the
   * field specified by the given index.
   * Returns <code>null</code> if the associated object of
   * this accessor is a primitive value or null reference.
   *
   * @param i the index of the field to be returned
   * @return an object accessor for the specified field's value
   * @exception java.rmi.RemoteException if an RMI failure occured.
   */
  public ObjectAccessor getField(int i) throws RemoteException;
}
