package de.renew.remote;

import java.rmi.*;
import java.rmi.server.*;
import java.io.Serializable;

import de.renew.simulator.Place;
import de.renew.simulator.PlaceInstance;
import de.renew.simulator.NetInstance;

import de.renew.event.PlaceEventListener;
import de.renew.event.PlaceEvent;
import de.renew.event.TokenEvent;

import collections.UpdatableSet;
import collections.HashedSet;
import collections.CollectionEnumeration;

/**
 * This class implements the <code>PlaceInstanceAccessor</code>
 * interface and nothing more than needed to implement it.
 * <p>

                        **********************
                        * UNDER CONSTRUCTION *
                        **********************

   missing methods: getMarking, addSerializableToken

 * </p>
 * PlaceInstanceAccessorImpl.java
 * Created: Sun Jul 16  2000
 * @author Michael Duvigneau
 */

public class PlaceInstanceAccessorImpl 
       extends UnicastRemoteObject
       implements PlaceEventListener, PlaceInstanceAccessor {
  
  /**
   * The place instance that is accessible by this Accessor.
   */
  private PlaceInstance placeInstance;

  /**
   * Creates a new place instance accessor for the given place
   * instance.
   *
   * @param placeInstance the place instance to access
   * @exception java.rmi.RemoteException if an RMI failure occured.
   */  
  public PlaceInstanceAccessorImpl(PlaceInstance placeInstance) throws RemoteException {
    super();
    this.placeInstance = placeInstance;
    placeInstance.addPlaceEventListener(this);
  }

  /* This method is specified by the Accessor interface. */
  public String asString() throws RemoteException {
    return placeInstance.toString();
  }

  /* This method is specified by the PlaceInstanceAccessor interface. */
  public int getID() throws RemoteException {
    return placeInstance.getPlace().getID();
  }

  /* This method is specified by the PlaceInstanceAccessor interface. */
  public PlaceAccessor getPlace() throws RemoteException {
    return new PlaceAccessorImpl(placeInstance.getPlace());
  }

  /* This method is specified by the PlaceInstanceAccessor interface. */
  public NetInstanceAccessor getNetInstance() throws RemoteException {
    return new NetInstanceAccessorImpl(placeInstance.getNetInstance());
  }

  /* This method is specified by the PlaceInstanceAccessor interface. */
  public MarkingAccessor getMarking() throws RemoteException {
    // TODO: implement this de.renew.remote.PlaceInstanceAccessor method
    return null;
  }

  /* This method is specified by the PlaceInstanceAccessor interface. */
  public void addSerializableToken(Serializable token) throws RemoteException {
    // TODO: implement this de.renew.remote.PlaceInstanceAccessor method
  }

  // ------------------------------------ Place event forwarding ----

  /**
   * The set of registered remote event listeners that want
   * to be informed about place events.
   */
  private UpdatableSet listeners = new HashedSet();

  /**
   * Registers the given remote event listener so that it will
   * receive future remote event messages concerning the accessed
   * place instance.
   * <p>
   * If the listener is registered already, the additional
   * registration try is ignored. The <code>equals()</code>
   * method serves as indicator to allow the mechanism to
   * work also for remote objects.
   * </p>
   * <p>
   * This method is specified by the <code>RemoteEventProducer</code>
   * interface which is required by the <code>PlaceInstanceAccessor</code>
   * interface.
   * </p>
   *
   * @param listener the listener to register
   * @exception java.rmi.RemoteException if a RMI failure occured.
   */
  public void addRemoteEventListener(RemoteEventListener listener) throws RemoteException {
    // There is no need to register this accessor as PlaceEventListener
    // because that was already done at construction time.
System.out.print("Including remote listener... ");
    listeners.include(listener);
System.out.println("now "+listeners.size()+" registered listeners.");
  }
  
  /**
   * Unregisters the given remote event listener so that it will
   * not receive future remote event messages from this place
   * instance accessor.
   * <p>
   * All listeners that equal the specified one will be
   * unregistered. If the listener was not registered, the
   * unregistration try is ignored.
   * </p>
   * <p>
   * This method is specified by the <code>RemoteEventProducer</code>
   * interface which is required by the <code>PlaceInstanceAccessor</code>
   * interface.
   * </p>
   *
   * @param listener the listener to unregister
   * @exception java.rmi.RemoteException if a RMI failure occured.
   */
  public void removeRemoteEventListener(RemoteEventListener listener) throws RemoteException {
System.out.print("Excluding remote listener... ");
    listeners.exclude(listener);
System.out.println("now "+listeners.size()+" registered listeners.");
  }

  /**
   * Converts the local place event into a remote event and
   * forwards it to all remote listeners.
   */
  public void markingChanged(PlaceEvent event) {
    fireRemoteEvent();
  }

  /**
   * Converts the local place event into a remote event and
   * forwards it to all remote listeners.
   */
  public void tokenAdded(TokenEvent event) {
    fireRemoteEvent();
  }

  /**
   * Converts the local place event into a remote event and
   * forwards it to all remote listeners.
   */
  public void tokenRemoved(TokenEvent event) {
    fireRemoteEvent();
  }


  /**
   * Converts the local place event into a remote event and
   * forwards it to all remote listeners.
   */
  public void tokenTested(TokenEvent event) {
    fireRemoteEvent();
  }

  /**
   * Converts the local place event into a remote event and
   * forwards it to all remote listeners.
   */
  public void tokenUntested(TokenEvent event) {
    fireRemoteEvent();
  }

  /**
   * The common reaction to all place events is to send a
   * remote event to all currently registered remote event
   * listeners.
   */
  private void fireRemoteEvent() {

    CollectionEnumeration listenerEnum;
    RemoteEventListener listener;

System.out.print("forwarding remote event to "+listeners.size()+" listeners...");
 
    synchronized(listeners) {
      listenerEnum = listeners.elements();
      while (listenerEnum.hasMoreElements()) {

System.out.print(" *");

	listener = (RemoteEventListener)listenerEnum.nextElement();
	try {
	  listener.update();
	} catch (RemoteException e) {
	  System.err.println("PlaceInstanceAccessor: Remote event to "
			     +listener+" probably got lost due to "+e);
	}
      }
    }

System.out.println(" Done.");
 
  }

}
