package de.renew.remote;

import java.rmi.*;
import java.rmi.server.*;

import de.renew.simulator.Transition;
import de.renew.simulator.TransitionInstance;
import de.renew.simulator.NetInstance;

import de.renew.event.TransitionEventListener;
import de.renew.event.FiringEvent;

import collections.UpdatableSet;
import collections.HashedSet;
import collections.CollectionEnumeration;

/**
 * This class implements the <code>TransitionInstanceAccessor</code>
 * interface and nothing more than needed to implement it.
 * <p>
 * </p>
 * TransitionInstanceAccessorImpl.java
 * Created: Mon Jul 17  2000
 * @author Michael Duvigneau
 */

public class TransitionInstanceAccessorImpl
       extends UnicastRemoteObject
       implements TransitionEventListener, TransitionInstanceAccessor {
  
  /**
   * The transition instance that is accessible by this Accessor.
   */
  private TransitionInstance transitionInstance;

  /**
   * Creates a new transition instance accessor for the given transition
   * instance.
   *
   * @param transitionInstance the transition instance to access
   * @exception java.rmi.RemoteException if an RMI failure occured.
   */  
  public TransitionInstanceAccessorImpl(TransitionInstance transitionInstance) throws RemoteException {
    super();
    this.transitionInstance = transitionInstance;
    transitionInstance.addTransitionEventListener(this);
  }

  /* This method is specified by the Accessor interface. */
  public String asString() throws RemoteException {
    return transitionInstance.toString();
  }

  /* This method is specified by the TransitionInstanceAccessor interface. */
  public int getID() throws RemoteException {
    return transitionInstance.getTransition().getID();
  }

  /* This method is specified by the TransitionInstanceAccessor interface. */
  public TransitionAccessor getTransition() throws RemoteException {
    return new TransitionAccessorImpl(transitionInstance.getTransition());
  }

  /* This method is specified by the TransitionInstanceAccessor interface. */
  public NetInstanceAccessor getNetInstance() throws RemoteException {
    return new NetInstanceAccessorImpl(transitionInstance.getNetInstance());
  }

  // ------------------------------- Transition event forwarding ----


  /**
   * The set of registered remote event listeners that want
   * to be informed about transition events.
   */
  private UpdatableSet listeners = new HashedSet();

  /**
   * Registers the given remote event listener so that it will
   * receive future remote event messages concerning the accessed
   * transition instance.
   * <p>
   * If the listener is registered already, the additional
   * registration try is ignored. The <code>equals()</code>
   * method serves as indicator to allow the mechanism to
   * work also for remote objects.
   * </p>
   * <p>
   * This method is specified by the <code>RemoteEventProducer</code>
   * interface which is required by the <code>TransitionInstanceAccessor</code>
   * interface.
   * </p>
   *
   * @param listener the listener to register
   * @exception java.rmi.RemoteException if a RMI failure occured.
   */
  public void addRemoteEventListener(RemoteEventListener listener) throws RemoteException {
    // There is no need to register this accessor as TransitionEventListener
    // because that was already done at construction time.
System.out.print("Including remote listener... ");
    listeners.include(listener);
System.out.println("now "+listeners.size()+" registered listeners.");
  }

  
  /**
   * Unregisters the given remote event listener so that it will
   * not receive future remote event messages from this transition
   * instance accessor.
   * <p>
   * All listeners that equal the specified one will be
   * unregistered. If the listener was not registered, the
   * unregistration try is ignored.
   * </p>
   * <p>
   * This method is specified by the <code>RemoteEventProducer</code>
   * interface which is required by the <code>TransitionInstanceAccessor</code>
   * interface.
   * </p>
   *
   * @param listener the listener to unregister
   * @exception java.rmi.RemoteException if a RMI failure occured.
   */
  public void removeRemoteEventListener(RemoteEventListener listener) throws RemoteException {
    // TODO: implement this de.renew.remote.RemoteEventProducer method
System.out.print("Excluding remote listener... ");
    listeners.exclude(listener);
System.out.println("now "+listeners.size()+" registered listeners.");
  }


  /**
   * Converts the local transition event into a remote event and
   * forwards it to all remote listeners.
   */
  public void firingStarted(FiringEvent event) {
    fireRemoteEvent();
  }


  /**
   * Converts the local transition event into a remote event and
   * forwards it to all remote listeners.
   */
  public void firingComplete(FiringEvent event) {
    fireRemoteEvent();
  }
  

  /**
   * The common reaction to all transition events is to send a
   * remote event to all currently registered remote event
   * listeners.
   */
  private void fireRemoteEvent() {

    CollectionEnumeration listenerEnum;
    RemoteEventListener listener;

System.out.print("forwarding remote event to "+listeners.size()+" listeners...");
 
    synchronized(listeners) {
      listenerEnum = listeners.elements();
      while (listenerEnum.hasMoreElements()) {

System.out.print(" *");

	listener = (RemoteEventListener)listenerEnum.nextElement();
	try {
	  listener.update();
	} catch (RemoteException e) {
	  System.err.println("TransitionInstanceAccessor: Remote event to "
			     +listener+" probably got lost due to "+e);
	}
      }
    }

System.out.println(" Done.");
 
  }

}
