package de.renew.simulator;

import collections.*;
import de.renew.unify.*;

public abstract class AssignBinder implements Binder {
    Variable variable;
    boolean checkBound;

    public AssignBinder(Variable variable, boolean checkBound) {
	this.variable=variable;
	this.checkBound=checkBound;
    }

    public int bindingBadness(Searcher searcher) {
	// The possibly incomplete value in the variable
	// will not be stored anywhere, so that it is
	// not required to make a copy of it.
	if (Unify.isBound(variable)) {
	    // This binder has become useless. It should
	    // be removed soon, but aborting the search is still
	    // preferable.
	    return 1;
	} else {
	    CollectionEnumeration enum=getCandidates(variable.getValue());
	    int result=BindingBadness.clip(enum.numberOfRemainingElements());  
	    return result;
	}
    }

    public void bind(Searcher searcher) {
	if (Unify.isBound(variable) && !checkBound) {
	    // The variable is already bound.
	    // No need to try different values.
	    // It is not even required to register a trigger.
	    // That will be done, if at all, by the binder
	    // that actually removes the token.
	    searcher.search();
	} else {
	    // Make sure that the trigger notifies
	    // the searchable if its state changes,
	    // because in that case the possible binding would have
	    // to be rechecked.
	    TriggerableCollection triggerables=getTriggerables();
	    if (triggerables!=null) {
	      // There is a trigger.
	      searcher.insertTriggerable(triggerables);
	    }

	    CollectionEnumeration values=new RandomEnumeration(
              getCandidates(variable.getValue()));

	    while (values.hasMoreElements() && !searcher.isCompleted()) {
		Object value=values.nextElement();

		int checkpoint=searcher.recorder.checkpoint();
		try {
		    Unify.unify(variable,value,searcher.recorder);
		    // Ok, unification succeeded.
		    searcher.search();
		}
		catch (Impossible e) {
		    // Unification failed.
		}
		searcher.recorder.restore(checkpoint);
	    }
	}
    }

  /**
   * Return a list of possible values. It is allowed to return values
   * that are impossible, but this will affect performance.
   * This list must be robust enough not to change internally
   * due to search activities.
   *
   * @param pattern a possibly incomplete pattern that the candidates
   *   should match
   * @return an enumeration of values
   */
  public abstract CollectionEnumeration getCandidates(Object pattern);

  /**
   * Get the triggerable collection where we can register
   * for updates of the candidate list.
   *
   * @return the triggerable collection or <tt>null</tt> if the
   *   enumeration will always be the same
   */
  public abstract TriggerableCollection getTriggerables();
}
