package de.renew.simulator;

import de.renew.util.Semaphor;
import java.lang.reflect.InvocationTargetException;

// Global synchronization order:
//   0th locally synchronized
//   1st lock SearchQueue
//   2nd lock threadLock
//   3rd lock finder

public class ConcurrentSimulator
  implements Runnable, Simulator, SearchQueueListener
{
  public static void main(String[] args) {
    if (args.length!=1) {
      System.out.println(
        "Usage: java de.renew.simulator.ConcurrentSimulator CLASS");
      System.exit(0);
    }

    // Create a net.
    Net net=null;
    try {
      Class clazz=Class.forName(args[0]); // No reload required.
      net=(Net)clazz.getMethod("instance",
	new Class[0]).invoke(null,new Object[0]);
    }
    catch (Exception e) {
      System.out.println(args[0]+
        " isn't a valid net class as it seems.");
      if (e instanceof InvocationTargetException) {
	((InvocationTargetException)e).getTargetException().printStackTrace();
      }
      e.printStackTrace();
      System.exit(1);
    }
    
    NetInstance primaryInstance=SimulatorHelper.fireInitially(net,
      new ConsoleTracer("Initially","(0) "),true);

    if (primaryInstance!=null) {
      Simulator simulator=new ConcurrentSimulator();
      simulator.startRun();
    } else {
      System.out.println("The net did not accept the init request.");
    }
  }

  private Searcher searcher=new Searcher();

  private Thread thread;
  private TracerFactory tracerFactory;

  // These field may only be accessed by a thread that
  // has grabbed the thread lock.
  private ExecuteFinder executeFinder;
  private AbortFinder abortFinder;

  public ConcurrentSimulator(TracerFactory tracerFactory) {
    this.tracerFactory=tracerFactory;

    setupFinders();

    thread=new Thread(this);
    // Run the simulation thread in the background.
    thread.setPriority(Thread.MIN_PRIORITY);
    // Start the thread.
    thread.start();
  }

  public ConcurrentSimulator() {
    this(new ConsoleTracerFactory(""));
  }

  private void setupFinders() {
    // Protect the finders against concurrent use.
    synchronized (threadLock) {
      executeFinder=new ExecuteFinder();
      abortFinder=new AbortFinder(executeFinder);
    }
  }

  // In the concurrent simulator, you never know if there might
  // be more enabled transitions. So I am always active.
  public boolean isActive() {
    return true;
  }

  // The following variables control the background thread.
  private long cycle=1L;
  // possible modes:
  // -1: termination
  //  0: stopped
  //  1: single step
  //  2: continous run
  private int desiredMode=0;
  private boolean running=true;

  private int stepStatusCode=0;

  private boolean searchQueueIsEmpty=false;
  private boolean alreadyRegistered=false;

  private Object threadLock=new Object();
  
  public void searchQueueNonempty() {
    synchronized(threadLock) {
      searchQueueIsEmpty=false;
      alreadyRegistered=false;
      // Notify threads that may be waiting for a nonempty search queue.
      threadLock.notifyAll();
    }
  }
  
  private void registerAtSearchQueue() {
    synchronized(SearchQueue.class) {
      // Now synchronize locally to access fields safely.
      synchronized(threadLock) {
	searchQueueIsEmpty=SearchQueue.isTotallyEmpty();
	if (searchQueueIsEmpty && !alreadyRegistered) {
	  SearchQueue.insertListener(this);
	  alreadyRegistered=true;
	}
      }
    }
  }

  // Terminate the simulation once and for all.
  // Do some final clean-up and exit the thread.
  public synchronized void terminateRun() {
    synchronized (threadLock) {
      // Set the desired mode. There is no need
      // to notify anyone, because that will be done by
      // the call to requestMode later on.
      desiredMode=-1;

      // Just in case a search is currently running, abort it
      // by notifying the finder.
      abortFinder.abortSearch();
      // Notify the search process about the new mode.
      requestMode(-1);
    }

  }

  // This method is not synchronized, so that it may
  // interrupt a run or step request.
  public void stopRun() {
    synchronized (threadLock) {
      // Set the desired mode. There is no need
      // to notify anyone, because that will be done by
      // the call to requestMode later on.
      desiredMode=0;

      // Just in case a search is currently running, abort it
      // by notifying the finder.
      abortFinder.abortSearch();
      // Notify the search process about the new mode.
      requestMode(0);
    }
  }


  // By synchronizing I can guarantee that no other
  // thread spoils the value in stepStatusCode.
  public synchronized int step() {
    // Require one more step.
    requestMode(1);

    return stepStatusCode;
  }

  public synchronized void startRun() {
    // Tell the thread to start running.
    requestMode(2);
  }

  // This method used to be synchronized, but there is
  // really no point to it. All accessed fields are
  // controlled by the thread, which will also make
  // sure the this method is not executed concurrently.
  private void requestMode(int mode) {
    synchronized(threadLock) {
      // If another thread has already requested a termination,
      // that request will not be taken back.
      if (desiredMode<0) {
	return;
      }
      // Set the mode.
      desiredMode=mode;
      // Set the running state.
      if (mode>0) {
	running=true;
      }
      // Get the thread out of the sleeping state, if it happens to be there.
      threadLock.notifyAll();
      // Now we must wait for the thread to stop running, if we do not
      // want to let it run indefinitely.
      if (mode<2) {
	while (running) {
	  try {
	    threadLock.wait();
	  }
	  catch (InterruptedException e) {
	    // This is expected.
	  }
	}
      }
    }
  }

  public void run() {
    while (desiredMode>=0) {
      // Wait until there is a reasonable chance that
      // a transition instance might be enabled.
      synchronized(threadLock) {
	while (searchQueueIsEmpty && desiredMode==2) {
	  try {
	    threadLock.wait();
	  }
	  catch (InterruptedException e) {
	    // This is expected.
	  }
	}
      }

      // We note that the search was stopped explicitly.
      // Whenever this is not the case, we will find out
      // later and correct the code accordingly.
      stepStatusCode=statusStopped;
	    
      // Try to get a transition instance from the search queue.
      // this might fail, because another thread might have emptied the
      // queue concurrently.
      Searchable searchable=SearchQueue.extract();
      while (searchable!=null && desiredMode>0) {
	// Create a new finder. Earlier during the execution another
	// thread might have placed a stop request that has been
	// processed in the mean time.
	//
	// While we are creating the new finder, another thread
	// might accesse the old finder and place a stop
	// request in it. We solve this problem by rechecking
	// the desired mode afterwards.
	setupFinders();

	// Do we still want to search for a binding?
	if (desiredMode>0) {
	  // Search for a binding. Use the abort finder, so that
	  // the search may be interrupted.
	  searcher.searchOnce(abortFinder,searchable,searchable);

	  // Grant other threads a chance to run.
	  // Only needed on systems without preemptive multitasking.
	  thread.yield();
	  
	  if (abortFinder.isCompleted()) {
	    // Ok, reinsert this searchable. This must be done if either
	    // the search was successful or if the search was interrupted.
	    SearchQueue.includeNow(searchable);
	  }

	  if (executeFinder.isCompleted()) {
	    // Execute the binding asynchronously.
	    executeFinder.execute(tracerFactory.getTracer("Synchronously",
	      String.valueOf(cycle++)),true);

	    // Provide a new finder so that the old finder does not
	    // forbid garbage collection.
	    setupFinders();
	    
	    // Make a note that we found a binding.
	    stepStatusCode=statusStepComplete;
	    
	    // Stop if in single step mode.
	    synchronized (threadLock) {
	      if (desiredMode==1) {
		desiredMode=0;
	      }
	    }
	  }

	  // Try to find another binding.
	  searchable=SearchQueue.extract();
	}
      }

      if (searchable!=null) {
	// I have to put back the searchable
	// object that I extracted, because
	// I have not yet found time to analyse it.
	SearchQueue.includeNow(searchable);
	// Forget the searchable. It might have to be garbage collected.
	searchable=null;
      }

      // If the queue is empty, let's register there. It's free.
      //
      // This is done even desiredMode is 0, because I would have to
      // lock threadLock to safely access desiredMode. But after locking
      // threadLock, it is forbidden to access SearchQueue.
      //
      // On the other hand, I must not lock SearchQueue all the time,
      // otherwise I might inhibit notifications from SearchQueue
      // that might keep me running.
      registerAtSearchQueue();
      
      // I must synchronize on the thread lock, otherwise
      // notifications might get lost.
      synchronized(threadLock) {
	// Did we fail to find any bindings in single step mode?
	if (desiredMode==1) {
	  // Request no further firings.
	  desiredMode=0;

	  // Leave a note that there are currently no enabled transitions.
	  stepStatusCode=statusCurrentlyDisabled;
	}
	
	// Did the user request a break?
	if (desiredMode<=0) {
	  // I am no longer running.
	  running=false;
	  
	  // I must send a notification, just in case a stopper waits.
	  // If there are multiple stoppers, I'll notify them all.
	  threadLock.notifyAll();
	}
	
	// Now we must stick here until someone allows us to proceed.
	while (desiredMode==0) {
	  try {
	    threadLock.wait();
	  }
	  catch (InterruptedException e) {
	    // This is expected.
	  }
	}
      }
    }
    // Ok, exit the simulator because someone requested its termination.
  }

  // No need to refresh. I do not cache variables.
  public void refresh() {}
}


