package de.renew.simulator;

import collections.*;
import de.renew.unify.*;
import de.renew.function.Executor;
import de.renew.util.Value;
import de.renew.util.Null;
import java.util.Vector;
import java.lang.reflect.Array;

class FlexibleArcBinder implements Binder {
  Variable tokenVar;
  FlexibleArc arc;
  TokenReserver reserver;
  PlaceInstance instance;
  /**
   * This is the flexible arc occurrence that created this binder and
   * wants to be informed about the tokens that are actually moved.
   * This information is passed via the field inTokens.
   *
   * @see FlexibleArcOccurrence#inTokens
   */
  FlexibleArcOccurrence occurrence;

  FlexibleArcBinder(FlexibleArcOccurrence occurrence,
		    Variable tokenVar, TokenReserver reserver,
		    PlaceInstance instance, FlexibleArc arc)
  {
    this.occurrence=occurrence;
    this.tokenVar=tokenVar;
    this.reserver=reserver;
    this.instance=instance;
    this.arc=arc;
  }

  public int bindingBadness(Searcher searcher) {
    if (Unify.isBound(tokenVar)) {
      return 1;
    } else {
      // We must not try to bind the variable.
      return BindingBadness.max;
    }
  }

  private void rememberSingleToken(Object tok) {
    occurrence.inTokens.addElement(Unify.copyBoundValue(tok));
  }

  public void bind(Searcher searcher) {
    if (!Unify.isBound(tokenVar)) {
      // This should not happen.
      throw new RuntimeException("Flexible arc binder was invoked for an "+
	"incomplete value.");
    }

    // The array is analysed locally and need not be copied.
    Object tokens=tokenVar.getValue();

    occurrence.inTokens=new Vector();

    if (tokens==null) {
      // Let's ignore this arc. Probably the null value is meant as
      // a substitute for an empty array.
    } else if (tokens.getClass().isArray()) {
      Class elementType=tokens.getClass().getComponentType();
      
      int n=Array.getLength(tokens);
      
      for (int i=0;i<n;i++) {
	Object tok=null;
	try {
	  // Extract a single token and copy it, so that
	  // backtracking does not hurt.
	  tok=Array.get(tokens,i);
	  
	  if (elementType.isPrimitive()) {
	    tok=new Value(tok);
	  }
	  
	  if (arc.forwardFunction!=null) {
	    Object convertedToken=arc.forwardFunction.function(tok);
	    
	    if (arc.backwardFunction!=null) {
	      Object restoredToken=
		arc.backwardFunction.function(convertedToken);
	      Unify.unify(tok,restoredToken,null);
	    }
	    
	    tok=convertedToken;
	  }
	  
	  rememberSingleToken(tok);
	}
	catch (Exception e) {
	  // Abort this search process.
	  // A conversion did not succeed.
	  // e.printStackTrace();
	  return;
	}
      }
    } else if (tokens instanceof List) {
      List current=(List)tokens;
      while (!current.isNull()) {
	rememberSingleToken(current.head());
	if (!(current.tail() instanceof List)) // stop if open/corrupted list
	  break; 
	current=(List)current.tail();
      }
    } else if (tokens instanceof Vector) {
      Vector vector=(Vector)tokens;
      for (int i=0;i<vector.size();i++) {
	rememberSingleToken(vector.elementAt(i));
      }
    }

    if (!arc.isOutputArc() && occurrence.inTokens.size()>0) {
      // Make sure that the place instance notifies
      // the searchable if its marking changes,
      // because in that case the possible binding would have
      // to be rechecked. This is only required for input
      // arcs, because output arcs cannot disable or enable
      // a transition, and only if a token is actually moved.
      searcher.insertTriggerable(instance.triggerables());
    }      

    // Now we start computation with side effects.
    boolean success=true;
    int maxOk=0;
    while (success && maxOk<occurrence.inTokens.size()) {
      try {
	success=reserver.removeToken(instance,
				     occurrence.inTokens.elementAt(maxOk),0);
	if (success) {
	  // Reservation was successful.
	  maxOk++;
	}
      }
      catch (Exception e) {
	e.printStackTrace();
	success=false;
      }
    }
    
    if (success) {
      // All reservations succeeded.
      searcher.search();
    }
    
    // Undo the successful reservations.
    while (maxOk>0) {
      maxOk--;
      reserver.unremoveToken(instance,
			     occurrence.inTokens.elementAt(maxOk),0);
    }

    // The set of tokens has become invalid.
    occurrence.inTokens=null;
  }
}
