package de.renew.simulator;

import collections.*;

/**
 * The net list class keeps track of all well known net instances.
 * Nets may be configured to register their instance automatically during
 * creation confirmation or individual nets may be registered explicitly.
 *
 * Currently, no special structure is given to this list.
 *
 * Accesses to the net list are sznchronized to forbid concurrent updates.
 * No special actions must be taken on the caller side.
 * No deadlocks are possible.
 */
public class NetList {
    private static UpdatableSet netInstances=new HashedSet();

    /**
     * This is an entirely static class.
     * No instance creation is allowed.
     */ 
    private NetList() {}

    /**
     * Register a net instance. If the net instance has already
     * been registered, ignore.
     *
     * @param instance the net instance to be registered.
     */
    public static synchronized void register(NetInstance instance) {
	netInstances.include(instance);
    }

    /**
     * Remove a net instance from the list. If the net instance is not
     * contained in the list, ignore.
     *
     * @param instance the net instance to be removed.
     */
    public static synchronized void remove(NetInstance instance) {
	netInstances.exclude(instance);
    }

    /**
     * Create a snapshot of the list of registered net instances.
     * The net instances are returned in a arbitrary order.
     * 
     * @return an array that references all well-known net instances
     */
    public static synchronized NetInstance[] getAll() {
        CollectionEnumeration enum=netInstances.elements();
	NetInstance[] result=new NetInstance[netInstances.size()];
	int i=0;
        while (enum.hasMoreElements()) {
	    result[i++]=(NetInstance)enum.nextElement();
	}
	return result;
    }
}
