package de.renew.simulator;

import de.renew.expression.*;
import collections.*;
import de.renew.event.*;
import de.renew.database.Transaction;
import de.renew.database.TransactionSource;
import de.renew.unify.*;
import de.renew.util.Lock;
import de.renew.util.Null;
import java.util.Enumeration;
import java.util.Vector;
import java.io.Serializable;
import java.io.IOException;
import de.renew.util.DelayedFieldOwner;
import de.renew.util.RenewObjectOutputStream;
import de.renew.util.RenewObjectInputStream;

/**
 * I will collect tokens for a single place instance. I also
 * keep count of the number of tokens that are currently being tested.
 *
 * Whenever my marking is changed I will notify my triggerables.
 *
 * Anybody who needs to make sure that my marking does not change should
 * lock on me. Synchronisation does not help!
 */

public class PlaceInstance 
  implements PlaceEventProducer, Serializable, DelayedFieldOwner {
    
  static final long serialVersionUID = -6409146722455348472L;

  private NetInstance netInstance;
  private Place place;
  
  private int freeTokenCount;
  private int testedTokenCount;

  /** 
   * This field is not really transient, but as we want
   * to cut down the serialization recursion depth, it
   * is serialized manually.
   **/
  transient private TokenBag freeTokens;

  /** 
   * This field is not really transient, but as we want
   * to cut down the serialization recursion depth, it
   * is serialized manually.
   **/
  transient private TestTokenBag testedTokens;

  /**
   * A hashtable where all tokens are registered
   * with their ID. The registry counts also the number
   * of times that a token has been removed from
   * the place instance where this change has not yet been
   * committed to the database.
   */
  private IDRegistry registry;

  /**
   * Here we store tokens that constitute the initial marking.
   * These tokens are computed in the constructor, but their
   * insertion must be traced later on and maybe they must even
   * be inserted into the place instance.
   **/
  private UpdatableSeq initTokens;

  /**
   * This flag denotes whether the initial tokens have been put
   * into the place instance.
   */
  private boolean earlyTokens;
  
  /** 
   * All elements contained in this index must also
   * appear in <code>testIndex</code>.
   *
   * This field is transient and will be recomputed
   * on deserialization.
   **/
  transient private TupleIndex freeIndex;

  /** 
   * Must contain at least all elements
   * contained in <code>freeIndex</code>.
   *
   * This field is transient and will be recomputed
   * on deserialization.
   **/
  transient private TupleIndex testIndex;

  /** 
   * This field is not really transient, but as we want
   * to cut down the serialization recursion depth, it
   * is serialized manually.
   *
   * It contains those objects that might have to be triggered once after
   * a change to the current marking.
   **/
  transient private TriggerableCollection triggerables;

  /**
   * The order in which locks are requested is governed by
   * this number, which is unique within a single simulation.
   *
   * lockOrder and lock were originally final, but to allow the
   * creation of new values on deserialization the modifier
   * had to be removed.
   */
  public transient long lockOrder;

  /**
   * The lock that control access to this place instance.
   */
  public transient Lock lock;

  /**
   * The listeners that want to be informed about every update
   * of the current marking.
   **/
  private transient UpdatableSet listeners=new HashedSet();

  /**
   * Create a new place instance for a given net instance
   * reflecting a certain place. If desired, the constructor
   * will already calculate in the initial marking.
   *
   * @param netInstance the owning net instance
   * @param place the semantic level place
   * @param wantInitialTokens true, if initial marking should be calculated
   */
  PlaceInstance(NetInstance netInstance, Place place,
		boolean wantInitialTokens) throws Impossible
  {
    this.netInstance=netInstance;
    this.place=place;

    lockOrder=de.renew.util.Orderer.getTicket();
    lock=new Lock();

    // Currently, there are no tokens available.
    freeTokens=new TokenBag();
    testedTokens=new TestTokenBag();
    freeTokenCount=0;
    testedTokenCount=0;

    registry=new IDRegistry();

    freeIndex=new TupleIndex();
    testIndex=new TupleIndex();

    triggerables=new TriggerableCollection();

    // Copy the early token flag. Maybe this flag changes
    // later on. (Well, it shouldn't, but let's be safe.)
    earlyTokens=netInstance.getNet().earlyTokens;

    // Let's calculate the initial bag of tokens.
    initTokens=new LinkedList();
    CollectionEnumeration inscriptions=place.inscriptions.elements();
    while (inscriptions.hasMoreElements()) {
      VariableMapper mapper=new VariableMapper();
      Variable thisVariable=mapper.map(new LocalVariable("this",false));
      Unify.unify(thisVariable,netInstance,null);

      TokenSource tokenSource=(TokenSource)inscriptions.nextElement();
      Object token=tokenSource.createToken(mapper);
      initTokens.insertFirst(Null.convert(token));

      if (earlyTokens) {
	// I am requested to insert the tokens now.
	internallyInsertToken(token,SearchQueue.getTime());
      }
    }
  }

  public String toString() {
    return netInstance.toString()+"."+place.toString();
  }

  public NetInstance getNetInstance() {
    return netInstance;
  }

  public Place getPlace() {
    return place;
  }

  // Access the triggerables. We do not lock the place instance during
  // accesses to the triggerables contrary to earlier implementations.
  // The triggerables have to take care of locking themselves.
  public TriggerableCollection triggerables() {
    return triggerables;
  }

  public void addPlaceEventListener(PlaceEventListener 
    listener)
  {
    synchronized(listeners) {
      listeners.include(listener);
    }
  }
  
  public void removePlaceEventListener(PlaceEventListener
    listener)
  {
    synchronized(listeners) {
      listeners.exclude(listener);
    }
  }

  public CollectionEnumeration getDistinctTokens() {
    return freeIndex.getAllElements();
  }

  public CollectionEnumeration getDistinctTokens(Object pattern) {
    return freeIndex.getPossibleMatches(pattern);
  }

  public CollectionEnumeration getDistinctTestableTokens() {
    return testIndex.getAllElements();
  }

  public CollectionEnumeration getDistinctTestableTokens(Object pattern) {
    return testIndex.getPossibleMatches(pattern);
  }

  public int getNumberOfTokens() {
    return freeTokenCount;
  }

  /**
   * Return the number of tokens that are currently tested
   * by a transition, but that reside in this place.
   * We do not count multiple tests on a single token as
   * multiple tokens.
   */
  public int getNumberOfTestedTokens() {
    return testedTokenCount;
  }

  public boolean isEmpty() {
    return getNumberOfTokens()+getNumberOfTestedTokens()==0;
  }

  boolean containsToken(Object token) {
    return freeTokens.includesAnytime(token);
  }

  public int getTokenCount(Object token) {
    return freeTokens.getMultiplicity(token);
  }

  public TimeSet getFreeTimeSet(Object token) {
    return freeTokens.getTimeSet(token);
  }

  // Returns the delay until a set of untested tokens
  // is available that matches the given delay times.
  public double computeEarliestTime(Object token, TimeSet times) {
    return freeTokens.computeEarliestTime(token,times);
  }

  // Here consider only already tested tokens.
  public boolean containsTestedToken(Object token) {
    return testedTokens.includesTested(token);
  }
  // Here we consider both kinds of tokens.
  boolean containsTestableToken(Object token) {
    // nulls will be converted by the called methods.
    return containsToken(token) || containsTestedToken(token);
  }

  // Make sure a token keeps its ID. Only while the
  // token is reserved or contained in the place
  // it is guaranteed that it keeps its ID.
  public void reserve(Object token) {
    // Increase the number of registrations.
    registry.reserve(token);
  }

  public String getTokenID(Object token) {
    return registry.getID(token);
  }

  public void unreserve(Object token) {
    // Decrease the number of registrations.
    registry.unreserve(token);
    possiblyDiscard(token);
  }

  private void possiblyDiscard(Object token) {
    // Is the token available?
    if (!freeTokens.includesAnytime(token) &&
	!testedTokens.includesTested(token))
    {
      // No. Deregister it, if possible.
      registry.possiblyDiscard(token);
    }
  }

  // Try to remove the token. If no token is available at
  // the given time, an exception is thrown.
  double removeToken(Object token, double delay) throws Impossible {
    lock.lock();

    double removedTime=0;
    try {
      removedTime=freeTokens.removeWithDelay(token,delay);
    }
    catch (Exception e) {
      throw new Impossible();
    }

    freeTokenCount--;

    if (!freeTokens.includesAnytime(token)) {
      freeIndex.remove(token);
      if (!testedTokens.includesTested(token)) {
	testIndex.remove(token);
      }
    }

    // All changes to the place instance state have been performed.
    // It is now safe to notify the current transaction,
    // the triggerables and the listeners (if any) of the change.
    //
    // Up to now the token has still kept its ID, because concurrent
    // accesses to the place instances are stopped by the lock.
    Transaction transaction=TransactionSource.get();
    transaction.removeToken(this,token);

    triggerables.proposeSearch();

    synchronized(listeners) {
      TokenEvent te=new TokenEvent(this,token);
      CollectionEnumeration enum=listeners.elements();
      while (enum.hasMoreElements()) {
	PlaceEventListener listener=(PlaceEventListener)enum.nextElement();
	listener.tokenRemoved(te);
      }
      // Allow the master place to distribute the event, too.
      place.tokenRemoved(te);
    }

    lock.unlock();

    return removedTime;
  }

  // Multiple tests on the same token do not remove
  // the token multiple times. Instead, the token is
  // removed once, and the number of removals is recorded
  // by a multiple insertion into the bag of tested tokens.
  double testToken(Object token) throws Impossible {
    lock.lock();

    boolean wantNotify=false;

    double removedTime=SearchQueue.getTime();
    // Do we need to remove a token explicitly?
    // It seems so. This does not really matter, because testing is
    // not really visible in the timed (an hence sequential)
    // formalism.
    if (!testedTokens.includesTested(token)) {
      try {
	// Remove a token that is already in the place.
	removedTime=freeTokens.removeWithDelay(token,0);
      }
      catch (Exception e) {
	throw new Impossible();
      }

      // Fix the token counters. Only required if a
      // token was actually converted from free to tested.
      testedTokenCount++;
      freeTokenCount--;

      if (!freeTokens.includesAnytime(token)) {
	freeIndex.remove(token);
      }

      // We must notify the listeners only if a free token has
      // become tested. If a token was already tested, the new
      // test simply delays the time span until a possible
      // activation.
      wantNotify=true;
    }

    testedTokens.addTested(token,removedTime);

    if (wantNotify) {
      triggerables.proposeSearch();

      synchronized(listeners) {
	TokenEvent te=new TokenEvent(this,token);
	CollectionEnumeration enum=listeners.elements();
	while (enum.hasMoreElements()) {
	  PlaceEventListener listener=(PlaceEventListener)enum.nextElement();
	  listener.tokenTested(te);
	}
	// Allow the master place to distribute the event, too.
	place.tokenTested(te);
      }
    }

    lock.unlock();

    return removedTime;
  }

  /**
   * This method removes all tokens from the place and
   * puts them into a vector that is provided as an argument.
   * The method should be used when the simulator is at rest
   * or at least the place is locked and no firings are
   * testing tokens. Otherwise some tested token may not
   * be removable.
   *
   * For each token, the original time stamp is recorded
   * in a second vector. The same number of elements
   * is added to each vector, one entry for each token in
   * the place. Multiple identical tokens result in multiple
   * identical elements in the vector.
   *
   * @param tokens vector where removed tokens will be placed
   * @param timeStamps vector where time stamps will be placed
   */
  public void extractAllTokens(Vector tokens, Vector timeStamps) {
    lock.lock();

    // In the following, I do not remove tested token.
    UpdatableSeq uniqueValues=new collections.LinkedList();
    uniqueValues.prependElements(getDistinctTokens());
    
    while (!uniqueValues.isEmpty()) {
      Object token=uniqueValues.first();
      uniqueValues.removeFirst();

      // The following loop might be optimized, but
      // currently this is the safest solution.
      while (containsToken(token)) {
	double time=0;
	try {
	  time=removeToken(token,Double.NEGATIVE_INFINITY);
	} catch (Impossible e) {
	  throw new RuntimeException("Token was not removable.");
	}
	tokens.addElement(token);
	timeStamps.addElement(new Double(time));
      }

      possiblyDiscard(token);
    }

    lock.unlock();
  }

  /**
   * This method inserts a token into the place instance and
   * and tries to assign the given ID to it. It is an error,
   * if the token is already registered in the place
   * instance iwth a different ID.
   *
   * This method should only be called when restoring a
   * simulation state from a database. Only in that case
   * there is any reason to assign a particular ID to a token.
   *
   * @param token the token to be inserted
   * @param id the token's ID
   * @param time the time stamp of the token
   * @exception java.lang.RuntimeException if the object was already
   *   registered with a different ID
   */
  public void insertTokenWithID(Object token, String id, double time) {
    lock.lock();
    registry.setID(token,id);
    internallyInsertToken(token,time);
    lock.unlock();
  }

  /**
   * This method inserts tokens into the place instance with notifications
   * and index updates, although it will not invoke the transaction
   * mechanism. 
   *
   * This method must be used with care. It should only be called
   * in those cases where an invocation of the transaction
   * mechanism would be absolutely inappropriate, e.g., within
   * the transaction mechanims itself or if the transaction
   * must be called explicitly or if a net instance is restored
   * from a database.
   *
   * This method will take care of locking the place instance
   * automatically. It is however, allowed that the current thread
   * has already locked it, if this is required due to 
   * deadlock prevention.
   *
   * @param token the token to be inserted
   * @param time the time stamp of the token
   */
  public void internallyInsertToken(Object token, double time) {
    // Acquire the lock to protect against concurrent updates.
    lock.lock();
    
    // Insert the tokens.
    if (!freeTokens.includesAnytime(token)) {
      freeIndex.insert(token);
      // Maybe there is an appropriate token being tested?
      if (!testedTokens.includesTested(token)) {
	// No, remember the new token.
	testIndex.insert(token);
      }
    }

    freeTokens.add(token,time);
    freeTokenCount++;

    // Now inform triggerables and listeners.
    // By now they might find out by chance, so we should better
    // tell them directly.
    triggerables.proposeSearch();

    synchronized(listeners) {
      TokenEvent te=new TokenEvent(this,token);
      CollectionEnumeration enum=listeners.elements();
      while (enum.hasMoreElements()) {
	PlaceEventListener listener=(PlaceEventListener)enum.nextElement();
	listener.tokenAdded(te);
      }
      // Allow the master place to distribute the event, too.
      place.tokenAdded(te);
    }

    // Free the lock.
    lock.unlock();
  }

  /**
   * This method records the effect of a token insertion
   * in a transaction object. If automatic insertion is requested,
   * then the token is actually inserted after the transaction has
   * been committed.
   *
   * @param token the token to be inserted
   * @param time the time stamp of the token
   * @param automaticInsertion true, if the transaction should care about
   *   the insertion of the token.
   */
  public void transactionInsertToken(Object token,
				     double time,
				     boolean automaticInsertion)
  {  
    Transaction transaction=TransactionSource.get();
    transaction.addToken(this,token,
			 time,automaticInsertion);
  }

  /**
   * Insert a token into the place instance at a specified time.
   * The token deposit is written to the currently active transaction.
   * Only after that transaction has been committed, the token
   * will occur in the place instance. If not transaction is active,
   * the deposit takes place immediately.
   *
   * This is the standard method call that applications should
   * use to get tokens into a net. It will not print trace messages,
   * but it will ultimately notify the listeners and triggerables
   * of the change.
   *
   * This method will take care of locking the place instance
   * automatically. It is however, allowed that the current thread
   * has already locked the place instance, if this is required
   * due to deadlock prevention.
   *
   * @param token the token to be inserted
   * @param time the time stamp of the token
   */
  public void insertToken(Object token, double time) {
    transactionInsertToken(token,time,true);
  }

  // Undo a testToken call. Once the last token is returned,
  // a token is reinserted into the bag of free tokens.
  //
  // The time is used as the time stamp of the returned token,
  // if the current test is the last. Otherwise, the argument is
  // ignored.
  void untestToken(Object token) {
    lock.lock();

    double time=testedTokens.removeTested(token);
    // No need to change the test index, because a tested token
    // will become free or an equivalent token will remain tested.

    if (!testedTokens.includesTested(token)) {
      if (!freeTokens.includesAnytime(token)) {
	freeIndex.insert(token);
      }

      // Reinsert the token as a free token at the given time.
      freeTokens.add(token,time);
      
      // Fix the token count.
      freeTokenCount++;
      testedTokenCount--;

      // We must notify the listeners only if a tested token has
      // become free. If a token is still tested, the 
      // untest simply shortens the time span until a possible
      // activation.
      triggerables.proposeSearch();

      synchronized(listeners) {
	TokenEvent te=new TokenEvent(this,token);
	CollectionEnumeration enum=listeners.elements();
	while (enum.hasMoreElements()) {
	  PlaceEventListener listener=(PlaceEventListener)enum.nextElement();
	  listener.tokenUntested(te);
	}
	// Allow the master place to distribute the event, too.
	place.tokenUntested(te);
      }
    }
    lock.unlock();
  }

  /**
   * Print trace messages for the insertion of initial tokens.
   *
   * @param tracer An object where trace messages might be
   *               printed. The current implementation does not
   *               print trace messages.
   */
  private void traceInitialTokens(Tracer tracer) {
    CollectionEnumeration enum=initTokens.elements();
    // Anything to do?
    while (enum.hasMoreElements()) {
      Object token=Null.convert(enum.nextElement());
      if (place.getTrace()) {
	tracer.trace("    Initializing "+token+
		     " into "+this);
      }
    }
  }

  /**
   * My net will call this method while its creation is confirmed.
   * I will notify the database about all tokens that I kept during
   * my initialisation, if the tokens must be available early.
   *
   * You must call this method at most once.
   *
   * In fact, this method is rather a kludge in the sense
   * that transaction support should not be activated
   * at all if a net requests early confirmation.
   */
  void earlyConfirmation() {
    if (earlyTokens) {
      CollectionEnumeration enum=initTokens.elements();
      // Record each token individually in the database.
      while (enum.hasMoreElements()) {
	Object token=Null.convert(enum.nextElement());
	double time=SearchQueue.getTime();
	// Do not keep a note to insert to tokens
	// at commit time. They were added already during the
	// creation of this net instance.
	transactionInsertToken(token,time,false);
      }
    }
  }

  /**
   * My net will call this method while its creation is traced.
   *
   * @param tracer An object where trace messages might be
   *               printed. The current implementation does not
   *               print trace messages.
   */
  void earlyConfirmationTrace(Tracer tracer) {
    if (earlyTokens) {
      traceInitialTokens(tracer);
    }
  }
  
  /**
   * My net will call this method if its creation is confirmed.
   * I will free all tokens that I kept during my initialisation.
   *
   * You must call this method at most once.
   *
   * @param tracer An object where trace messages might be
   *               printed.
   */
  void lateConfirmation(Tracer tracer) {
    if (!earlyTokens) {
      traceInitialTokens(tracer);
	
      CollectionEnumeration enum=initTokens.elements();
      // Anything to do?
      while (enum.hasMoreElements()) {
	Object token=Null.convert(enum.nextElement());
	insertToken(token,SearchQueue.getTime());
      }
    }
    
    // Null the field to allow garbage collection.
    initTokens=null;
  }

  /**
   * Serialization method, behaves like default writeObject
   * method except storing the not-really-transient fields
   * freeTokens, testedTokens and triggerables.
   * If the stream used is a RenewObjectOutputStream, these
   * fields are delayed to cut down recursion depth.
   * The domain trace feature of this special stream is also
   * used.
   * @see de.renew.util.RenewObjectOutputStream
   **/
  private void writeObject(java.io.ObjectOutputStream out) throws IOException {
    RenewObjectOutputStream rOut = null;
    if (out instanceof RenewObjectOutputStream)
      rOut = (RenewObjectOutputStream)out;
    if (rOut != null) rOut.beginDomain(this);
    out.defaultWriteObject();
    if (rOut != null) {
      rOut.delayedWriteObject(freeTokens, this);
      rOut.delayedWriteObject(testedTokens, this);
      rOut.delayedWriteObject(triggerables, this);
      rOut.endDomain(this);
    } else {
      out.writeObject(freeTokens);
      out.writeObject(testedTokens);
      out.writeObject(triggerables);
    }
  }

  /**
   * Deserialization method, behaves like default readObject
   * method except restoring additional transient fields.
   * Creates new lock object and gets a new lock order ticket.
   * The method also restores the not-really-transient fields
   * freeTokens, testedTokens and triggerables, <b>if</b> the
   * stream used is <b>not</b> a RenewObjectInputStream.
   * In that case it also recomputes the transient fields
   * freeIndex and testedIndex.
   **/
  private void readObject(java.io.ObjectInputStream in)
               throws IOException, ClassNotFoundException {
    in.defaultReadObject();
    lockOrder=de.renew.util.Orderer.getTicket();
    lock=new Lock();
    listeners=new HashedSet();

    if (in instanceof RenewObjectInputStream) {
      // Do nothing, the fields will be
      // reassigned by the stream soon.
    } else {
      // Free tokens must be written first, so that
      // the indexes can be recomputed after reading the
      // tested tokens.
      freeTokens = (TokenBag)in.readObject();
      testedTokens = (TestTokenBag)in.readObject();
      triggerables = (TriggerableCollection)in.readObject();
      recomputeIndexes();
    }
  }

  /** 
   * Method used on deserialization by RenewObjectInputStream.
   * Reassigns values to the not-really-transient fields
   * freeTokens, testedTokens and triggerables, one at a
   * time.
   * Also recomputes freeIndex and testIndex when the
   * corresponding TokenBag gets reassigned.
   **/
  public void reassignField(Object value) throws java.io.IOException {
    if (value instanceof TokenBag) {
	freeTokens = (TokenBag)value;
    } else if (value instanceof TestTokenBag) {
      testedTokens = (TestTokenBag)value;
      recomputeIndexes();
    } else if (value instanceof TriggerableCollection) {
      triggerables = (TriggerableCollection)value;
    } else {
	throw new java.io.NotSerializableException(
		  "Value of unexpected type given to reassign().");
    }
  }

  /**
   * Rebuilds both TupleIndex fields (freeIndex and 
   * testIndex) from scratch, using the elements of
   * the TokenBag fields freeTokens and testedTokens.
   * Needed by the deserialization methods readObject()
   * and reassignField() to update the transient
   * TupleIndex fields.
   **/
  private void recomputeIndexes() {
      freeIndex=new TupleIndex();
      addToIndex(freeIndex,freeTokens.uniqueElements());
      testIndex=new TupleIndex();
      addToIndex(testIndex,freeTokens.uniqueElements());
      addToIndex(testIndex,testedTokens.uniqueElements());
  }

  /**
   * Inserts all elements of the given TokenBag into
   * the given TupleIndex. Used by recomputeIndexes().
   **/
  private void addToIndex(TupleIndex index, Enumeration tokenList) {
      while (tokenList.hasMoreElements()) {
	  Object token=tokenList.nextElement();
	  index.insert(token);
      }
  }
    
}

