package de.renew.simulator;

import collections.*;
import de.renew.event.*;
import de.renew.unify.Impossible;
import de.renew.unify.Variable;

import java.io.Serializable;
import java.io.IOException;
import java.io.NotSerializableException;
import de.renew.util.DelayedFieldOwner;
import de.renew.util.RenewObjectOutputStream;
import de.renew.util.RenewObjectInputStream;

/**
 * I represent an instance of a transition. Unlike a naked transition
 * I am tied to a net instance. I provide my name to the outside world
 * and keep track of those objects who like to be notified if I become
 * enabled.
 *
 * Other objects will register me as a listener of objects that might
 * enable me. I would like to be registered as a listener with
 * all objects that might enable me, but this is beyond my control.
 * So take a little care when you use me.
 *
 * I used to send out messages in the case that I might become
 * enabled. This functionality was removed, because no one
 * really ever needed it.
 *
 * @author Olaf Kummer
 **/

public class TransitionInstance implements Searchable, Triggerable,
  UplinkProvider, TransitionEventProducer, Serializable, DelayedFieldOwner
{

  static final long serialVersionUID = -2146903280470015837L;

  /**
   * My net instance to which I belong.
   **/
  NetInstance netInstance;
  /**
   * The transition that serves as a pattern for me.
   **/
  Transition transition;

  /**
   * The triggers that might notify me.
   * This field is not really transient, but as we want
   * to cut down the serialization recursion depth, it
   * is serialized manually.
   **/
  private transient TriggerCollection triggers=new TriggerCollection(this);

  /**
   * All pending firing events that have started, but not yet finished.
   **/
  private transient UpdatableSet pendingEvents=new HashedSet();

  /**
   * The listeners that I must notify.
   **/
  private transient UpdatableSet listeners=new HashedSet();

  /**
   * I (a transition instance) am created. I remember
   * the neccessary information.
   *
   * @param netInstance
   *   the net instance that needs to have an instance of some transition
   * @param transition
   *   the transition that must be instantiated
   **/
  TransitionInstance(NetInstance netInstance,
		     Transition transition)
  {
    this.netInstance=netInstance;
    this.transition=transition;
  }

  /**
   * I will use the name of my net instance plus the name of
   * my transition to build my own name.
   *
   * @return my name
   **/
  public String toString() {
    return netInstance.toString()+"."+transition.toString();
  }

  public NetInstance getNetInstance() {
    return netInstance;
  }

  public Transition getTransition() {
    return transition;
  }

  public TriggerCollection triggers() {
    return triggers;
  }

  public synchronized void addTransitionEventListener(TransitionEventListener 
    listener)
  {
    listeners.include(listener);
  }

  public synchronized void removeTransitionEventListener(TransitionEventListener
    listener)
  {
    listeners.exclude(listener);
  }

  /**
   * Somebody will call this method to tell me that I might
   * be enabled. I will tell the search queue about it, if I am
   * spontaneous. I will also tell my listeners.
   */
  public synchronized void proposeSearch() {
    // Notify the searcher, if this transition is
    // spontaneous. Otherwise, the notification was sporadic.
    // Maybe I should report an error instead of silently
    // ignoring the notification.
    if (transition.isSpontaneous()) {
      SearchQueue.includeNow(this);
    }
  }

  /**
   * Informs all TransitionEventListeners and the master transition
   * that this instance has just started firing.
   *
   * This method is called by FiringStartExecutables created
   * by occurrences of this transition instance.
   **/
  synchronized void firingStarted(FiringEvent fe) {
    // Make note for future updates.
    pendingEvents.include(fe);

    // Notify all listeners.
    CollectionEnumeration enum=listeners.elements();
    while (enum.hasMoreElements()) {
      TransitionEventListener listener=(TransitionEventListener)enum.nextElement();
      listener.firingStarted(fe);
    }    

    // Notify the master transition and all its listeners.
    transition.firingStarted(fe);
  }

  /**
   * Informs all TransitionEventListeners and the master transition
   * that this instance has just completed firing.
   *
   * This method is called by FiringCompleteExecutables created
   * by occurrences of this transition instance.
   **/
  synchronized void firingComplete(FiringEvent fe) {
    // No more need to reference this event.
    pendingEvents.exclude(fe);

    // Notify all listeners.
    CollectionEnumeration enum=listeners.elements();
    while (enum.hasMoreElements()) {
      TransitionEventListener listener=(TransitionEventListener)enum.nextElement();
      listener.firingComplete(fe);
    }    

    // Notify the master transition and all its listeners.
    transition.firingComplete(fe);
  }
  
  /**
   * Return all firing events for firings that have started,
   * but not yet completed.
   *
   * Before calling this method, it is necessary to synchronize on
   * this object. Only after the processing of all firing events
   * the synchronization may end.
   *
   * If the enumeration is required for a very long time,
   * it is suggested to copy it into an array.
   *
   * @see de.renew.event.FiringEvent
   *
   * @return an enumeration of firing events.
   */
  public CollectionEnumeration pendingFiringEvents() {
    return pendingEvents.elements();
  }

  /**
   * My net will call this method if it creation is confirmed.
   * Just to be safe, I will register myself at the search queue
   * if I am spontaneous.
   **/
  void createConfirmation() {
    if (transition.isSpontaneous()) {
      SearchQueue.includeNow(this);
    }
  }

  public boolean listensToChannel(String channel) {
    return !transition.isSpontaneous() && 
      transition.uplink.name.equals(channel);
  }

  private void createAndSearchOccurrence(Variable params, Searcher searcher) {
    int checkpoint=searcher.recorder.checkpoint();
    
    try {
      // Create a new occurrence. If there is an uplink,
      // synchronize using the parameters.
      TransitionOccurrence occurrence=
	new TransitionOccurrence(this,params,searcher);

      // No exception was thrown, so this transition
      // may be checked.
      searcher.addOccurrence(occurrence);
      searcher.addBinders(occurrence.getBinders());
      searcher.search();
      searcher.removeBinders(occurrence.getBinders());
      searcher.removeOccurrence(occurrence);
    }
    catch (Impossible e) {
      // I had to refuse the search request completely.
      // Probably there is a cyclic dependency of some
      // calculations. This should really be detected during
      // the syntax check.
    }

    searcher.recorder.restore(checkpoint);
  }
  /**
   * This method is called if somebody wants me to continue a
   * search for an activated binding by matching my uplink
   * with a downlink whose parameters are given as arguments.
   *
   * @param params
   *   the variable that holds the value that it send over the channel
   * @param searcher
   *   the searcher that is supposed to actually perform the search.
   **/
  public void bindChannel(Variable params, Searcher searcher) {
    // This method does exactly what the method createAndSearchOccurrence
    // would do. I hesitate to join the the methods, because 
    // this method is not supposed to handle spontaneous transitions.
    createAndSearchOccurrence(params,searcher);
  }

  /**
   * This method is called if somebody wants me to start a
   * search for an activated binding. I initiate the search and
   * let the searcher do the rest.
   *
   * @param searcher
   *   the searcher that is supposed to actually perform the search.
   **/
  public void startSearch(Searcher searcher) {
    createAndSearchOccurrence(null,searcher);
    // Is the next line really needed?
    searcher.recorder.restore(); // no bindings sensible
  }


  /**
   * Serialization method, behaves like default writeObject
   * method except checking an additional error condition
   * and writing the not-really-transient field triggers.
   * If the stream used is a RenewObjectOutputStream, the
   * field is delayed to cut down recursion depth and the
   * domain trace feature will be used.
   * @see de.renew.util.RenewObjectOutputStream
   *
   * Throws an NotSerializableException if the object cannot
   * be serialized because there are pending events or 
   * bindings.
   */
  private void writeObject(java.io.ObjectOutputStream out) throws IOException {
    RenewObjectOutputStream rOut = null;
    if (out instanceof RenewObjectOutputStream)
      rOut = (RenewObjectOutputStream)out;
    if (rOut != null) rOut.beginDomain(this);
    if (pendingEvents.isEmpty()) {
	out.defaultWriteObject();
	if (rOut != null) {
	  rOut.delayedWriteObject(triggers, this);
	} else {
	  out.writeObject(triggers);
	}
    } else {
      throw new NotSerializableException("Active bindings "
		+ " at transition instance " + this);
    }
    if (rOut != null) rOut.endDomain(this);
  }

  /** 
   * Deserialization method, behaves like default readObject
   * method except restoring the not-really-transient field
   * triggers, <b>if</b> the stream used is <b>not</b> a
   * RenewObjectInputStream.
   * It also assigns default values to some transient fields.
   **/
  private void readObject(java.io.ObjectInputStream in)
      throws IOException, ClassNotFoundException {
    in.defaultReadObject();
    pendingEvents=new HashedSet();
    listeners=new HashedSet();

    if (in instanceof RenewObjectInputStream) {
      // Do nothing, the field will be 
      // reassigned by the stream soon.
    } else {
      triggers = (TriggerCollection)in.readObject();
    }
  }


  /** 
   * Method used on deserialization by RenewObjectInputStream.
   * Reassigns values to the not-really-transient fields
   * triggers, one at a time.
   **/
  public void reassignField(Object value) throws java.io.IOException {
    if (value instanceof TriggerCollection) {
      triggers = (TriggerCollection)value;
    } else {
	throw new java.io.NotSerializableException(
		  "Value of unexpected type given to reassign().");
    }
  }

}
