package de.renew.simulator;

import de.renew.expression.*;
import collections.*;
import java.util.Enumeration;
import de.renew.unify.*;
import de.renew.event.FiringEvent;

/**
 * I represent a transition occurrence, i.e. a transition instance
 * that is (or might be) firing. A transition instance that fires
 * simultaneously with itself will require multiple transition
 * occurrences. Each occurence has its own set of local variables,
 * its own set of binders, and its own set of inscription
 * occurrences.
 *
 * @author Olaf Kummer
 **/
public class TransitionOccurrence implements Occurrence {
  /**
   * I am an occurence of this transition instance.
   * The instance determines the transition (hence my semantics)
   * and the net instance (hence my sphere of influence).
   **/
  TransitionInstance transitionInstance;
  /**
   * Here I store my local variables. They are distinct from
   * other occurrences.
   **/
  VariableMapper mapper;

  private UpdatableSeq inscriptionOccurrences;
  private UpdatableSeq binders;

  /**
   * I (a transition occurrence) am created. If required, my inscriptions
   * will be instantiated, my expressions will be evaluated, and
   * a list of binders will be prepared for the searcher.
   *
   * @param transitionInstance
   *   the transition instance of which I will be an occurence
   * @param params
   *   parameters of the synchronous channel that invoked me,
   *   if any
   * @param searcher
   *   the searcher that initiated the occurence of the
   *   transition
   * @exception de.renew.unify.Impossible
   *   I cannot possibly occur
   *   with the current parameters, regardless of the marking
   *   of the neighboring places. (E.g. two parameters in the channel
   *   are required to be the same, but aren't.)
   **/
  TransitionOccurrence(TransitionInstance transitionInstance, Variable params,
    Searcher searcher) throws Impossible
  {
    int checkpoint=searcher.recorder.checkpoint();

    try {
      this.transitionInstance=transitionInstance;
      
      // Make a simple variable mapper.
      mapper=new VariableMapper();
      
      // Make the special variable "this" known to the mapper. 
      Variable thisVariable=mapper.map(new LocalVariable("this",false));
      try {
	Unify.unify(thisVariable,transitionInstance.netInstance,null);
      }
      catch (Impossible e) {
	throw new RuntimeException(
          "Unification failed unexpectedly.");
      }

      Transition transition=transitionInstance.transition;
       
      if (!transition.isSpontaneous()) {
	Object up=transition.uplink.params.startEvaluation(mapper,
	  searcher.recorder,searcher.calcChecker);
	Unify.unify(up,params,searcher.recorder);
	// If the unification produced an exception,
	// then no bindings are possible.
      }

      inscriptionOccurrences=new LinkedList();
      
      CollectionEnumeration enum=transition.inscriptions.elements();
      while (enum.hasMoreElements()) {
	TransitionInscription inscription=
	  (TransitionInscription)enum.nextElement();
	inscriptionOccurrences.prependElements(
	  inscription.makeOccurrences(mapper,transitionInstance.netInstance,
	    searcher));
      }

      binders=new LinkedList();

      enum=inscriptionOccurrences.elements();
      while (enum.hasMoreElements()) {
	binders.prependElements(
	  ((InscriptionOccurrence)enum.nextElement()).makeBinders(searcher));
      }
    }
    catch (Impossible e) {
      // Undo bindings.
      searcher.recorder.restore(checkpoint);
      // Rethrow the exception;
      throw e;
    }
  }

  /**
   * Return all the binders that I calculated during creation.
   *
   * @return an enumeration of binders
   *
   * @see de.renew.simulator.Binder
   **/
  CollectionEnumeration getBinders() {
    return binders.elements();
  }
  
  /**
   * This provides access to my variable mapper.
   *
   * @return a variable mapper
   **/
  public VariableMapper getMapper() {
    return mapper;
  }

  /**
   * This provides access to the transition instance that makes me occur.
   *
   * @return an the transition instance that makes me occur
   **/
  public TransitionInstance getTransitionInstance() {
    return transitionInstance;
  }

  /**
   * I will provide executables that
   * perform the required actions according to the
   * current bindings of variables. This includes, among other
   * actions, token moves and tests.
   *
   * All my binders must have been processed before it is allowed
   * to call this method.
   *
   * @param copier
   *   a copier that I can use to rescue the current values of
   *   variables from backtracking
   * @return an enumeration of executables
   *
   * @see de.renew.simulator.Executable
   **/
  public CollectionEnumeration makeExecutables(Copier copier) {
    // Make a copy of all local variables.
    VariableMapper copiedMapper=mapper.makeCopy(copier);

    // We can assume that inscriptionOccurrences is not null.
    UpdatableSeq executables=new LinkedList();
    // Possibly insert a trace object.
    if (transitionInstance.transition.getTrace()) {
      executables.insertFirst(new TraceExecutable(
        "  Firing "+transitionInstance));
    }

    CollectionEnumeration enum=inscriptionOccurrences.elements();
    while (enum.hasMoreElements()) {
      InscriptionOccurrence inscrOcc=
        (InscriptionOccurrence)enum.nextElement();

      CollectionEnumeration additionalExecutables=
	inscrOcc.makeExecutables(copier,copiedMapper);
      executables.prependElements(additionalExecutables);
    }

    // No longer print a trace message, because we must reorder
    // the executables. But this tracing will work fine if there is
    // no reordering.
    //
    //    executables.insertFirst(new TraceExecutable(
    //      "  Firing "+transitionInstance));

    // Insert notifiers that can report the variable mapping.
    // Both will report the same firing event to the
    // listeners. Event identities are unique so that
    // the listener can associate start and end of a
    // firing.
    FiringEvent event=new FiringEvent(transitionInstance,copiedMapper);
    executables.insertFirst(new FiringStartExecutable(event));
    executables.insertFirst(new FiringCompleteExecutable(event));
    
    return executables.elements();
  }

  public String toString() {
    return transitionInstance.toString();
  }

}
