package de.renew.util;

import java.io.*;
import java.util.*;
import java.util.zip.*;

public class SelectiveClassLoader extends ClassLoader {

  private Hashtable cache = new Hashtable();

  private String[] classSources;
  private boolean[] isZipFile;
  private ZipFile[] zipFiles;
  private String[] classSelectors;
  private boolean[] wantInclusion;
  private boolean wantFallBack=false;

  private boolean mutable;

  public SelectiveClassLoader() {
      this(System.getProperty("java.class.path"));
  }

  public SelectiveClassLoader(String path) {
      this(path, new String[0]);
  }

  public SelectiveClassLoader(String path, String[] selectors) {
    mutable=true;
    
    setClassPath(path);
    setSelectors(selectors);
  }

  public void setClassPath(String path) {
    setClassPath(StringUtil.splitPaths(path));
  }
  
  public synchronized void setClassPath(String[] sources) {
    if (!mutable) {
      throw new RuntimeException("Class loader already locked.");
    }

    classSources=sources;

    isZipFile=new boolean[classSources.length];
    zipFiles=new ZipFile[classSources.length];
    for (int i=0;i<classSources.length;i++) {
      String ext=StringUtil.getExtension(classSources[i]);
      isZipFile[i]="jar".equals(ext) ||	"zip".equals(ext);
      if (isZipFile[i]) {
	try {
	    zipFiles[i]=new ZipFile(classSources[i]);
	} catch (IOException e) {
	    zipFiles[i]=null;
	}
      } else if (!classSources[i].endsWith(File.separator)) {
	  classSources[i]=classSources[i]+File.separator;
      }
      System.out.println("SelectiveClassLoader class source: "+classSources[i]);
    }
  }

  // Selectors: separated list of class name prefixes,
  // preceeded by a minus when requested for exclusion.
  // Left entries take precedence over right entries.
  // If nothing matches, classes are loaded. If the default
  // should be not to load a class, append a minus as last entry.
  //
  // E.g.:
  //   -some.where.Name:some.where.:this.too.:-
  // loads all classes in some.where and this.too and all
  // subpackages, but not some.where.Name and nothing outside
  // of these classes.
  public void setSelectors(String selectors) {
    setSelectors(StringUtil.splitPaths(selectors));
  }
  
  public synchronized void setSelectors(String[] selectors) {
    if (!mutable) {
      throw new RuntimeException("Class loader already locked.");
    }

    classSelectors=selectors;

    wantInclusion=new boolean[classSelectors.length];
    for (int i=0;i<classSelectors.length;i++) {
      wantInclusion[i]=!classSelectors[i].startsWith("-");
      if (!wantInclusion[i]) {
	classSelectors[i]=classSelectors[i].substring(1);
      }
    }
  }

  public void setFallBack(boolean wantFallBack) {
    if (!mutable) {
      throw new RuntimeException("Class loader already locked.");
    }

    this.wantFallBack=wantFallBack;
  }

  public void lock() {
    mutable=false;
  }

  private boolean isSelectedClass(String name) {
    for (int i=classSelectors.length-1;i>=0;i--) {
      if (name.startsWith(classSelectors[i])) {
	return wantInclusion[i];
      }
    }
    // Nothing matched.
    return true;
  }

  private byte[] tryLoadZipFile(ZipFile file, String name) throws Exception {
    name=name.replace('.','/')+".class";
    ZipEntry entry=file.getEntry(name);
    byte[] data=new byte[(int)entry.getSize()];
    InputStream stream=file.getInputStream(entry);
    stream.read(data);
    stream.close();

    return data;
  }
  
  private byte[] tryLoadNormFile(String path, String name) throws Exception {
    name=path+StringUtil.classToFile(name);
    InputStream stream=new FileInputStream(name);
    byte[] result=new byte[stream.available()];
    stream.read(result);
    stream.close();
   
    return result;
  }

  private byte[] loadClassData(String name) {
    for (int i=0;i<classSources.length;i++) {
	//System.err.println("Trying class source "+classSources[i]);
      byte[] result=null;
      try {
	if (isZipFile[i]) {
	  result=tryLoadZipFile(zipFiles[i],name);
	} else {
	  result=tryLoadNormFile(classSources[i],name);
	}
      }
      catch (Exception e) {
	// The file could not be loaded here. Ignore this path entry.
      }
      if (result!=null) {
	return result;
      }
    }

    // Class not found.
    return null;
  }

  public synchronized Class loadClass(String name, boolean resolve)
    throws ClassNotFoundException
  {
    Class c=(Class)cache.get(name);

    if (c==null) {
	//System.out.println("Trying to load class "+name+"...");
      if (isSelectedClass(name)) {
	//System.out.println("Trying to reload class "+name+"...");
	try {
	  byte data[]=loadClassData(name);
	  c=defineClass(name,data,0,data.length);
	}
	catch (Exception e) {
	  // The class was not found. We might retry using the
	  // system class loader. But do we want to?
	  if (wantFallBack) {
	    c=findSystemClass(name);
	  } else {
	    throw new ClassNotFoundException(name);
	  }
	}
	catch (ClassFormatError err) {
	    // this may happen due to file system's upper/lower conversions
	    throw new ClassNotFoundException(name);
	}
      } else {
	// The class is not intended to be loaded by this
	// class loader.
	c=findSystemClass(name);
      }
      cache.put(name, c);
    }

    if (resolve)
      resolveClass(c);

    return c;
  }
}
